/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.client;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.ClientApplication;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.pua.prot.manager.ProtocolRelatedOperation;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.PidProtocol;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.util.concurrent.TimeUnit;

/**
 * Empfänger für protokoll-bezogene Antwortdatensätze von Pua. Der erste Antwortdatensatz von Pua enthält die Protokoll-Id, über die die Kommunikation ab diesem
 * Punkt abgewickelt wird.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class ProtocolAnswerReceiver extends AnswerReceiver {

	/** Debug Ausgabe */
    private static final Debug _debug = Debug.getLogger();

	/** Zeigt an ob die Protokoll-Id bereits empfangen wurde. */
	private boolean _receivedProtocolId = false;

	/**
	 * <code>true</code> wenn der Receiver verwendet werden soll. (In diesem Fall erwartet der Empfänger als erstes Packet eine positive Quittung mit
	 * Protokoll-Id.
	 */
	private boolean _createProtocolRequest;
	
	private long protocolId = -1;
	
	// true wenn eine job-id oder eine Fehlermeldung empfangen wurde
	private boolean _receivedFirstTelegram = false;

	/**
	 * Erzeugt Senke für protokoll-bezogene Datensätze. Meldet sich sofort auf protokoll-bezogene Antwortdaten von PuA an.
	 *
	 * @param dav                   Verbindung zum Datenverteiler.
	 * @param client                Empfängerobjekt.
	 * @param requestId             Anfrage-Id.
	 * @param createProtocolRequest <code>true</code> wenn der Receiver zum Empfang eines Protokolls verwendet werden soll. (In diesem Fall erwartet der Empfänger
	 *                              als erstes Packet eine positive Quittung mit Protokoll-Id.
	 * @param timeout               Maximale Wartezeit, in der eine Antwort von der SWE Protokolle und Auswertungen eingehen muss.{@code 0} bedeutet: Unbegrenzte
	 *                              Wartezeit. Achtung: Werden nur Online-Daten abgefragt, die selten auftreten (&gt; timeout), oder müssen Archivdaten wieder
	 *                              eingelagert werden, so kann es zu einem Timeout kommen, obwohl 'Protokolle und Auswertungen' arbeitet.
	 *
	 * @throws ConfigurationException Fehler beim Anmelden fü rDatenempfang.
	 */
	public ProtocolAnswerReceiver(ClientDavInterface dav, ClientApplication client, long requestId, boolean createProtocolRequest, long timeout)
			throws ConfigurationException {
		super(dav, client, PidProtocol.atgProtocolRequest, PidProtocol.aspAnswer, client, requestId, timeout);
		this._createProtocolRequest = createProtocolRequest;
	}

	@Override
	protected void update(ResultData result) {
		Data data = result.getData();
		if(data.getUnscaledValue(PidProtocol.opCode).longValue() == ProtocolRelatedOperation.PROTOCOL_COMPLETION_INTERFACE){
			// werden von de.bsvrz.pua.prot.client.ProtocolCompletionAnswerReceiver verarbeitet und nicht hier.
			return;
		}
		try {
			if(_receivedProtocolId) {
				_debug.finest("Empfange weiteren Datensatz");
				//überprüfen ob korrekte id und korrektes senderobjekt verwendet wird.
				if(receiver.equals(data.getReferenceValue(PidProtocol.sender).getSystemObject())
						&& (data.getUnscaledValue(PidProtocol.protocolId).longValue() == requestId
						|| data.getUnscaledValue(PidProtocol.protocolId).longValue() == protocolId)) {
					// Es wird sowohl die Request-Id als auch die Protokoll-Id geprüft, da alte PuA-Server Fehlermelungen mit der falschen ID senden
					// (Eigentlich richtig wäre hier die Protokoll-Id)
					incoming.push(result);
				}
			}
			else {
				_debug.finest("Empfange ersten Datensatz");
				//überprüfen ob korrekte id und senderobjekt übertragen und protokoll-Id im Datenteil 
				if(receiver.equals(data.getReferenceValue(PidProtocol.sender).getSystemObject())
				   && data.getUnscaledValue(PidProtocol.protocolId).longValue() == requestId) {
					//prüfen ob protokoll-id im datenteil.
					Long pid;
					if(data.getUnscaledValue(PidProtocol.opCode).longValue() == ProtocolRelatedOperation.ACK_FLAG && _createProtocolRequest) {
						try {
							pid = SerializerUtil.deserializeId(data.getUnscaledArray(PidProtocol.data).getByteArray());
							protocolId = pid;
							_receivedProtocolId = true;
						}
						catch(Exception e) {
							_debug.error(ErrorMessage.UNEXPECTED_DATA_TYPE, e);
							incoming.push(result);
						}
					}
					else {
						//sonst abspeichern
						incoming.push(result);
					}
					synchronized(this){
						_receivedFirstTelegram = true;
						notifyAll();
					}
				}
			}
		}
		catch(Exception e) {
			_debug.warning(ErrorMessage.UNEXPECTED_DATA_TYPE + data, e);
		}
	}

	/**
	 * Liefert die Auftrags-Id.
	 *
	 * @return Auftrags-Id oder <code>null</code> falls die Auftrags-Id unbekannt ist. Bei Protokoll-Erstellen-Aufträgen ist die Auftrags-ID
	 * identisch zur Protokoll-Id.
	 */
	public Long getJobId() {
		if(_receivedProtocolId) {
			return protocolId;
		}
		else {
			return null;
		}
	}

	/**
	 * Liefert die Auftrags-Id.
	 *
	 * @return Auftrags-Id oder <code>null</code> falls die Auftrags-Id unbekannt ist.
	 *
	 * @deprecated Diese Methode liefert nicht mehr die Protokoll-Id, sondern die eindeutige ID des protokollbezogenen Auftrags
	 */
	@Deprecated
	public final Long getProtocolId() {
		return getJobId();
	}

	public void waitForAnswer(long timeoutMillis) throws InterruptedException, FailureException {
		synchronized(this){
			long startTime = time();
			while(!_receivedFirstTelegram){
				wait(timeoutMillis);
				long endTime = time();
				timeoutMillis -= endTime - startTime;
				startTime = endTime;
				if(timeoutMillis <= 0){
					throw new FailureException(ErrorMessage.TIMEOUT, LoggerException.WARNING);
				}
			}
		}
	}

	private long time() {
		return TimeUnit.NANOSECONDS.toMillis(System.nanoTime());
	}
}
