/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009-2018 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import antlr.Token;
import com.google.common.base.Joiner;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.pua.prot.aggregations.Aggregation;
import de.bsvrz.pua.prot.aggregations.Average;
import de.bsvrz.pua.prot.aggregations.Maximum;
import de.bsvrz.pua.prot.aggregations.Minimum;
import de.bsvrz.pua.prot.aggregations.Sum;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.MemberCheck;
import de.bsvrz.pua.prot.util.ProcessingParameter;
import de.bsvrz.pua.prot.util.Qualifier;
import de.bsvrz.pua.prot.util.attributes.DurationAttributeDescription;
import de.bsvrz.pua.prot.util.attributes.TempAttributeDescription;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.time.*;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeFormatterBuilder;
import java.time.format.DateTimeParseException;
import java.time.format.ResolverStyle;
import java.util.*;

/**
 * Symboltabelle.<br> Die Symboltabelle wird von {@link de.bsvrz.pua.prot.interpreter.ProtocolParser} während des Parsens aufgebaut.<br> Standardwerte aus dem
 * Skript (aus dem Bereich Standards) werden nur übernommen, wenn sie keine Eingabeparameter des Benutzers überschreiben.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class SymbolTable {

	private static final Debug _debug = Debug.getLogger();
	
	private static final DateTimeFormatter DATE_FORMAT = new DateTimeFormatterBuilder().appendPattern("dd.MM.uuuu").toFormatter(Locale.GERMANY).withResolverStyle(ResolverStyle.STRICT);
	private static final DateTimeFormatter TIME_FORMAT = new DateTimeFormatterBuilder().appendPattern("HH:mm:ss,SSS").toFormatter(Locale.GERMANY).withResolverStyle(ResolverStyle.STRICT);
	/** Alias Definitionen ( = Spaltennamen) */
	final HashMap<String, String> _aliases;

	/**
	 * Enthält die temporären Attribute. Diese Hashtable wird mit Werten gefüllt, bevor irgendein anderer Zugriff auf die temporären Objekte stattfindet. Daher
	 * kann (unter der Voraussetzung dass alle temporären Objekte im 'Spalten'-Bereich verwendet werden müssen) davon ausgegangen werden, dass für jedes temporäres
	 * Objekt ein Eintrag in diese Liste erfolgt.
	 */
	private final List<TempAttributeDescription> _tempAttributes;

	/** Zeitdauer Attribut <code>null</code> falls das Zeitdauer Attribut nicht im Skript verwendet wird. */
	DurationAttributeDescription _durationAttribute = null;

	/** Definierte Aggregationsfunktionen <String, Aggregation> = <Name der Aggregation, Aggregationsobjekt> */
	final Map<String, Class<? extends Aggregation>> _aggregationFcts;

	/** Pseudoobjekt Definitionen (optional).<br> [x,y]: x: Alias, y: Pids (jedem Alias kann genau eine Pidliste zugeordnet sein */
	final Map<String, String[]> _pseudoObjects;

	/** Bindungen der freien Aspekte<br> [String, String] = [Platzhalter, Aspektname] */
	final Map<String, String> _aspectBindings;

	// Ab hier kommen die Standardbindungen / Infos aus ProcessingParameter

	/** Zeitbereiche, die berücksichtigt werden sollen */
	final List<Tuple<Long, Long>> _periodList;

	private SystemObject _defaultObject;

	/** Simulationsvariante */
	short _simVar;

	/** True falls das Protokoll gespeichert werden soll */
	boolean _saveProtocol;

	/** Hauptobjekt */
	final SystemObject _mainObject;

	/** Enthält den gewünschten Protokolltyp, der von der Anfrage kommt. True: Änderungsprotokoll, False: Zustandsprotokoll, Null: Standardwert im Skript verwenden */
	private final Boolean _deltaProtocol;

	/** Objekt zur Überprüfung der PIDs etc */
	private MemberCheck _check;

	/** Datenmodel */
	private final DataModel _model;

	/** Eingabeparameter des Auftraggebers */
	final ProcessingParameter _pp;

	/** Prüft, ob es schon eine Pseudoobjekt-Definition gibt. Falls ja, und eine zweite angelegt wird, wird ein Fehler geworfen */
	protected boolean _pseudoObjectDefinitionInitialized;

	private final Collection<String> _possibleAspectBindings = new HashSet<>();

	/**
	 * Erzeugen der Symboltabelle
	 *
	 * @param model Datenmodell
	 * @param pp    Eingabeparameter vom Auftraggeber, <code>null</code> wenn das Skript nur Interpretiert (auf güligkeit geprüft) aber nicht ausgeführt werden soll
	 *
	 * @throws CriticalParserException Falls es zu einem Kommunikationsfehler kommt
	 */
	public SymbolTable(final DataModel model, final ProcessingParameter pp) throws CriticalParserException {
		_check = null;
		this._model = model;
		_aliases = new LinkedHashMap<>();
		_tempAttributes = new ArrayList<>();
		_aggregationFcts = new LinkedHashMap<>();
		_aggregationFcts.put("maximum", Maximum.class);
		_aggregationFcts.put("minimum", Minimum.class);
		_aggregationFcts.put("summe", Sum.class);
		_aggregationFcts.put("mittelwert", Average.class);

		this._pp = pp;
		
		if(pp != null) {
			_mainObject = pp.getMainObject();

			if(pp.getPeriodList() != null) {
				_periodList = new ArrayList<>(pp.getPeriodList());
			}
			else {
				_periodList = new ArrayList<>();
			}

			_simVar = pp.getSimVar();
			_saveProtocol = pp.getSaveProtocol();
			_deltaProtocol = pp.getDeltaProtocol();

			_pseudoObjects = new LinkedHashMap<>(pp.getPseudoObjects());
			_aspectBindings = new LinkedHashMap<>(pp.getAspBindings());
		}
		else {
			_mainObject = null;
			_periodList = new ArrayList<>();
			_simVar = Qualifier.INVALID_SIMVAR;
			_saveProtocol = false;
			_deltaProtocol = null;

			_pseudoObjects = new LinkedHashMap<>();
			_aspectBindings = new LinkedHashMap<>();
		}
	}


	/**
	 * Festlegen welches Objekt die Typüberprüfungen durchführt.
	 *
	 * @param check Objekt das die Typüberprüfungen durchführen soll.
	 */
	public void setCheck(final MemberCheck check) {
		this._check = check;
	}

	/**
	 * Fügt ein Objekt hinzu, für das das Protokoll erstellt werden soll. Fügt das Objekt nur hinzu, wenn noch kein Objekt des gleichen Objekttyps hinterlegt
	 * wurde.
	 *
	 * @param so Objekt das Hinzugefügt werden soll
	 *
	 * @throws CriticalParserException Falls es zu einem Kommuniktaionsfehler kommt
	 */
	public void addStandardObject(final SystemObject so) throws CriticalParserException {
		if(so != null) {
			if(_defaultObject == null) {
				_defaultObject = so;
			}
			else {
				// Hier nur beim Skript erstellen Fehler werfen.
				// Wenn das Skript ausgeführt wird, ist es normal, das Skript-Standardobjekt zu ignorieren, wenn bereits eines per Parameter angegeben wurde
				throw new CriticalParserException(
						ErrorMessage.MULTIPLE_OBJECTS + _defaultObject + "; " + so
				);
			}
		}
	}

	/**
	 * Gibt das Objekt zurück, durch das der übergebene Typ ersetzt werden soll. Im Skriptquelltext kommen nur Objekttypen vor. Diese sind durch konkrete Objekte
	 * zu ersetzen.
	 *
	 * @param sot Objekttyp, für den ein Objekt zurückgeliefert werden soll.
	 *
	 * @return Objekt das statt des übergebenen Objekttyps im Protokoll verwendet werden soll. Objekttyp, falls keine Bindung für den Objekttyp besteht.
	 */
	public SystemObject getObject(final SystemObjectType sot) {
		if(sot == null) {
			return null;
		}
		SystemObject mainObject = getMainObject();
		if(mainObject != null && mainObject.isOfType(sot)) {
			return mainObject;
		}
		return sot;
	}

	public SystemObject getMainObject() {
		return _mainObject == null ? _defaultObject : _mainObject;
	}

	/**
	 * Fügt eine Alias-Definition in die Symboltabelle ein
	 *
	 * @param token Token. Wird für die Fehlerausgabe verwendet.
	 * @param name  Name des Alias
	 * @param obj   Referenzierter Objekttyp
	 *
	 * @throws CriticalParserException Falls der Alias bereits definiert wurde oder die Definition fehlerhaft ist.
	 */
	public void addAliasDefinition(final Token token, final String name, final String obj) throws CriticalParserException {
		if(name != null && obj != null) {
			if(_aliases.containsKey(name)) {
				throw new SemanticErrorException(InterpreterErrorMessage.REDEFINITION_ALIAS + name);
			}
			else {
				// überprüfen, ob es sich dabei um einen Objekttyp handelt
				if(_check == null) _check = new MemberCheck(_model, this);
				if(_check.whatIs(obj) != MemberCheck.OBJECTTYPE) throw new SemanticErrorException(InterpreterErrorMessage.OBJECTTYPE_EXPECTED + obj, token);
				_aliases.put(name, obj);
			}
		}
	}

	/**
	 * Liefert die dem Alias zugeordneten Objekt-PIDs zurück
	 *
	 * @param alias Name des Alias
	 *
	 * @return Eine Liste mit Objekten, die dem Alias durch die Pseudoobjektdefinition zugeordnet wurde. null, falls dem Alias keine Aspekte zugeordnet wurden
	 *
	 * @throws SemanticErrorException Falls dem Alias keine Objekte zugeordnet wurden oder es kein Alias ist.
	 */
	public String[] getAliasObjects(final String alias) throws SemanticErrorException {
		String objs[] = null;
		if(alias != null && isAlias(alias)) objs = _pseudoObjects.get(alias);

		return objs;
	}

	/**
	 * Fügt ein temporäres Attribut in die Symboltabelle ein (Aus Bereich 'Spalte').
	 *
	 * @param name       Name des temporären Attributs
	 * @param columnName Spaltenname
	 *
	 * @throws SemanticErrorException Der Name des temporären Attributs entspricht {@link DurationAttributeDescription#DURATION}
	 */
	public void addTempAttribute(final String name, final String columnName) throws SemanticErrorException {
		if(name != null) {
			if(name.equals(DurationAttributeDescription.DURATION)) {
				throw new SemanticErrorException(InterpreterErrorMessage.DURATION_USED_AS_NAME);
			}
			else {
				/*
				 * Es erfolgt keine Prüfung, ob ein Attribut dieses Namens bereits vorhanden, da ein temporäres Attribut
				 * durchaus mehrfach im Spalten-Bereich verwendet werden darf. Es muss jedoch ausgeschlossen werden,
				 * dass ein temporäres Attribut ebenfalls "Zeitdauer" genannt wird.
				 */
				_tempAttributes.add(new TempAttributeDescription(name, columnName));
			}
		}
	}

	public boolean isTempAttributeDefined(final String name) {
		for(final TempAttributeDescription tempAttribute : _tempAttributes) {
			if(tempAttribute.getName().equals(name)) return true;
		}
		return false;
	}

	public TempAttributeDescription getTempAttributeDefinition(final String name) {
		for(final TempAttributeDescription tempAttribute : _tempAttributes) {
			if(tempAttribute.getName().equals(name)) return tempAttribute;
		}
		return null;
	}
	
	public TempAttributeDescription getOrCreateTempAttributeDefinition(final String name) {
		for(final TempAttributeDescription tempAttribute : _tempAttributes) {
			if(tempAttribute.getName().equals(name)) return tempAttribute;
		}
		TempAttributeDescription newAttr = new TempAttributeDescription(name, null);
		_tempAttributes.add(newAttr);
		return newAttr;
	}

	public List<TempAttributeDescription> getTempAttributes() {
		return Collections.unmodifiableList(_tempAttributes);
	}

	/**
	 * Fügt eine Pseudoobjekt-Definition zur Symboltabelle hinzu.<br> Erlaubt die teiweise Übernahme einer Pseudoobjektdefinition: Falls einem Alias schon eine
	 * PID-Liste zugeordnet ist, dann wird nur dieser Alias übersprungen, die übrigen werden aber eingefügt.
	 *
	 * @param token      Token (nur für die Angabe von Zeilennummern in den Fehlermeldungen)
	 * @param aliasList  Liste mit Aliasen
	 * @param objectList Liste der Objekte, die einem Alias zugeordnet sind. (enthält ""-Werte für Platzhalter)
	 *
	 * @throws CriticalParserException Bei Kommunikationsfehler oder falls in der Pseudoobjektdefinition ungültige Pids verwendet werden
	 */
	public void addPseudoObjectDefinition(final Token token, final List<String> aliasList, final List<String> objectList) throws CriticalParserException {
		try {
			if(_pseudoObjectDefinitionInitialized) throw new SemanticErrorException(InterpreterErrorMessage.MULTIPLE_PSEUDO_OBJECT_SPECIFICATION);
			_pseudoObjectDefinitionInitialized = true;
			if(aliasList != null && objectList != null) {
				// Anzahl der Objekte muss Vielfaches der Aliase sein
				if(objectList.size() % aliasList.size() == 0) {
					int i = 0;
					for(final String alias : aliasList) {
						// überprüfen, ob es schon ein Mapping für diese Objekttypen gibt:
						if(!_pseudoObjects.containsKey(alias)) {
							// Jetzt die zugeordneten Objekte zusammenstellen
							final String[] concreteObjects = new String[objectList.size() / aliasList.size()];
							for(int j = 0; j < concreteObjects.length; j++) {
								concreteObjects[j] = objectList.get(j * aliasList.size() + i);
							}

							// überprüfen, ob gültige pids verwendet wurden:
							try {
								for(final String pid : concreteObjects) {
									if(!pid.isEmpty() && _model.getObject(pid) == null) throw new SemanticErrorException(InterpreterErrorMessage.OBJECT_EXPECTED + pid);
								}
							}
							catch(ConfigurationException ce) {
								throw new CriticalParserException(ErrorMessage.COMMUNICATION);
							}

							_pseudoObjects.put(alias, concreteObjects);
						}
						i++;
					}
				}
				else {
					throw new SemanticErrorException(
							InterpreterErrorMessage.TOO_MANY_TOO_LESS_ALIAS + "Angegeben: " + aliasList.size() + ", benötigt: " + objectList.size(), token
					);
				}
			}
			else {
				throw new SemanticErrorException(InterpreterErrorMessage.NO_ALIAS_OR_OBJECT, token);
			}
		}
		catch(CriticalParserException cpe) {
			cpe.line = token.getLine();
			throw cpe;
		}
	}

	/**
	 * Fügt eine Bindung eines Aspekts in die Symboltabelle ein, falls es für den angegebenen Platzhalter noch keine Bindung gibt. Gibt es schon eine Bindung für
	 * den Platzhalter, so wird die neue Bindung nicht eingetragen.
	 *
	 * @param placeholder Platzhalter des Aspekts (z.B. '?1')
	 * @param aspectName  Name des Aspekts
	 */
	public void addStandardBinding(final String placeholder, final String aspectName) {
		if(placeholder != null && aspectName != null) {
			if(!_aspectBindings.containsKey(placeholder)) _aspectBindings.put(placeholder, aspectName);
		}
	}

	/**
	 * Liefert gebundenen Aspekt zurück
	 *
	 * @param placeholder Platzhalter des Aspekts
	 *
	 * @return Aspekt, der an den Platzhalter gebunden ist. Falls keine Bindung existiert, wird null zurückgegeben.
	 */
	public String getBinding(final String placeholder) {
		if(placeholder != null) {
			_possibleAspectBindings.add(placeholder);
			return _aspectBindings.get(placeholder);
		}
		else {
			return null;
		}
	}

	/**
	 * Fügt die Definition einer Aggregationsfunktion in die Symboltabelle ein
	 *
	 * @param t               Token: Für Zeilenangabe im Fehler.
	 * @param aggregationName Name der Aggregation
	 * @param className       Klassenname der Aggregation
	 *
	 * @throws SemanticErrorException Falls ein ungültiger Klassenname angegeben wurde
	 */
	public void addAggregationDefinition(final Token t, final String aggregationName, final String className) throws SemanticErrorException {
		if(aggregationName == null) {
			throw new SemanticErrorException(InterpreterErrorMessage.NO_AGGREGATION_NAME, t);
		}
		else if(!_aggregationFcts.containsKey(aggregationName)) {
			final Class<?> cl;
			try {
				cl = Class.forName(className);
			}
			catch(ClassNotFoundException ignored) {
				throw new SemanticErrorException(InterpreterErrorMessage.WRONG_CLASSNAME + className, t);
			}
			if(cl.isInterface()) {
				throw new SemanticErrorException(InterpreterErrorMessage.CANT_CREATE_AGGREGATIONOBJECT + cl.getName());
			}
			try {
				final Class<? extends Aggregation> aggregationClass = cl.asSubclass(Aggregation.class);
				_aggregationFcts.put(aggregationName, aggregationClass);
			}
			catch(ClassCastException ignored) {
				throw new SemanticErrorException(
						InterpreterErrorMessage.WRONG_CLASSTYPE + cl.getName() + " ist nicht vom Typ " + Aggregation.class.getName(), t
				);
			}
		}
	}

	/**
	 * Stellt den Standard Anfragebereich ein, sofern der Benutzer keinen angegeben hat. Hat der Benutzer bereits Zeitintervalle definiert, werden diese nicht
	 * geändert.
	 *
	 * @param startDate Startdatum
	 * @param startTime Startzeit
	 * @param endDate   Enddatum
	 * @param endTime   Ende des Zeitraums
	 *
	 * @throws SemanticErrorException Falls es sich um einen fehlerhaften Zeitbereich handelt
	 */
	public void addStandardPeriod(final String startDate, final String startTime, final String endDate, final String endTime) throws SemanticErrorException {
		if(_periodList.isEmpty()) {
			long start, end;
			start = getTimeInMillis(startDate, startTime);
			end = getTimeInMillis(endDate, endTime);
			if(start > end) {
				final long temp = end;
				end = start;
				start = temp;
			}
			_periodList.add(new Tuple<>(start, end));
		}
	}

	/**
	 * Liefert die Zeit in Millisekunden zurück und überprüft Datum- & Zeitangabe auf Korrektheit
	 *
	 * @param date Datum
	 * @param time Zeit
	 *
	 * @return Millisekunden seit EPOCH
	 *
	 * @throws SemanticErrorException Falls Zeit oder Datum nicht korrekt angegeben wurden.
	 */
	private static long getTimeInMillis(final String date, final String time) throws SemanticErrorException {
		try {
			StringBuilder dateStr = new StringBuilder(date);
			StringBuilder timeStr = new StringBuilder(time);
			final List<String> warnings = new ArrayList<>();
			assumeChar(dateStr, 2, '.', warnings);
			assumeChar(dateStr, 5, '.', warnings);
			assumeChar(timeStr, 2, ':', warnings);
			assumeChar(timeStr, 5, ':', warnings);
			assumeChar(timeStr, 8, ',', warnings);
			
			LocalDate localDate = LocalDate.parse(dateStr, DATE_FORMAT);
			LocalTime localTime = LocalTime.parse(timeStr, TIME_FORMAT);
			
			long result = LocalDateTime.of(localDate, localTime).atZone(ZoneId.of("Europe/Berlin")).toInstant().toEpochMilli();

			if(!warnings.isEmpty()) {
				_debug.warning("Ignoriere Fehler in ungültigem Datum:", Joiner.on("\n").join(warnings));
			}

			if(result < 0) {
				throw new SemanticErrorException(InterpreterErrorMessage.INVALID_YEAR + localDate.getYear());
			}

			return result;
		}
		catch(DateTimeParseException ex) {
			throw new SemanticErrorException(InterpreterErrorMessage.DATE_OR_TIME_EXPECTED + ex.getMessage());
		}
	}

	private static void assumeChar(final StringBuilder builder, final int index, final char c, final List<String> warnings) {
		if(index < builder.length() && builder.charAt(index) != c) {
			warnings.add("Erwarte das Zeichen '" + c + "' an Position " + (index + 1) + ". Es war aber: '" + builder.charAt(index) + "'.");
			builder.setCharAt(index, c);
		}
	}

	/**
	 * Überprüft, ob es einen Alias diesen Namens gibt.
	 *
	 * @param name Name des Alias
	 *
	 * @return True, falls name der Name eines Alias ist.
	 */
	public boolean isAlias(final String name) {
		if(name != null) {
			return _aliases.containsKey(name);
		}
		else {
			return false;
		}
	}

	/**
	 * Gibt den Alias mit diesem Namen zurück
	 *
	 * @param name Name des Alias
	 *
	 * @return Den Alias oder null falls er nicht existiert
	 */
	public String getAlias(final String name) {
		if(name != null) {
			return _aliases.get(name);
		}
		else {
			return null;
		}
	}

	/**
	 * Zeitdauer Attribut hinzufügen.
	 *
	 * @param columnName Spaltenname des Zeitdauer Attributs.
	 */
	public void addDurationAttribute(final String columnName) {
		if(columnName == null) {
			_durationAttribute = new DurationAttributeDescription();
		}
		else {
			_durationAttribute = new DurationAttributeDescription(columnName);
		}
	}

	/** Gibt gewünschten Protokolltyp zurück, der von der Protokollerstellungs-Anfrage kommt. True: Änderungsprotokoll, False: Zustandsprotokoll, Null: Standardwert im Skript verwenden
	 * @return true, false oder null.
	 */
	public Boolean getDeltaProtocol() {
		return _deltaProtocol;
	}

	/**
	 * Gibt alle im Skript definierten ungebundenen oder gebundenen Aspekte zurück.
	 * @return Liste mit Aspekt-Namen
	 */
	public List<String> getPossibleAspectBindings() {
		return new ArrayList<>(_possibleAspectBindings);
	}
}
