/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.manager;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.manager.datamanager.DataManagerInterface;
import de.bsvrz.pua.prot.manager.requests.DeleteScriptRequest;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.kernsoftware.ConnectionManager;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.util.*;

/**
 * Objekte dieser Klasse nehmen Objekte vom Datenverteiler entgegen. Das Objekt wird dann an den {@link de.bsvrz.pua.prot.manager.datamanager.DataManager}
 * weiterleitet. Überwacht Änderungen an der Menge der Objekte. Wird der Menge ein Objekt hinzugefügt, so meldet sich der Empfänger auf dieses Objekt an. Wird
 * ein Skriptobjekt aus der Menge entfernt, dann meldet dich der Receiver auch wieder von diesem Objekt ab.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public final class ScriptRequestReceiver implements ClientReceiverInterface, MutableSetChangeListener {

	/** Daten Manager. Wird beim Eintreffen von Daten benachrichtigt. */
	private final DataManagerInterface _dataManager;

	/** ATG & APS der zu überwachenden Daten. */
	private final DataDescription _dataDescription;

	/** Empfangsoptionen. Delta oder Normal. */
	private final ReceiveOptions _options;

	/** Empfänger oder Senke. */
	private final ReceiverRole _role;

	/** Verbindung zum Datenverteiler. */
	private final ClientDavInterface _dav;

	/** Debug-Ausgabe. */
	private static final Debug _debug = Debug.getLogger();

	/** Set, das auf Änderungen überwacht wird. */
	private final MutableSet _mutableSet;

	/** Zeigt ob Anmeldung auf alle Skriptobjekte durchgeführt wurde. */
	private boolean _ready = false;

	/**
	 * @param dav             Verbindung zum Datenverteiler
	 * @param dataManager     Daten Manager. Wird beim Eintreffen von Daten benachrichtigt.
	 * @param mutableSet      Zu überwachende ObjektMenge.
	 * @param dataDescription Atg und Aspekt.
	 * @param options         Empfangsoptionen. Delta oder Normal.
	 * @param role            Empfänger oder Senke
	 *
	 * @throws FailureException Fehler bei der Anmeldung oder beim Zugriff auf die Objektmenge.
	 */
	public ScriptRequestReceiver(
			ClientDavInterface dav,
			DataManagerInterface dataManager,
			MutableSet mutableSet,
			DataDescription dataDescription,
			ReceiveOptions options,
			ReceiverRole role) throws FailureException {
		this._dataDescription = dataDescription;
		this._dataManager = dataManager;
		this._options = options;
		this._role = role;
		this._dav = dav;
		this._mutableSet = mutableSet;

		mutableSet.addChangeListener(this);
		// -> Seit KSv3.0 wird durch addChangeListener die Methode update(MutableSet,...) nicht mehr automatisch bei der Anmeldung aufgerufen
		// -> Manueller Aufruf notwenig!
		//Bis KSv2.0 wurde die update Methode mit mit allen im Set enthaltenen Objekten aufgerufen.
		update(mutableSet, mutableSet.getElements().toArray(new SystemObject[mutableSet.getElements().size()]), new SystemObject[0]);
	}


	/**
	 * Daten von Objekt trifft ein
	 *
	 * @see de.bsvrz.dav.daf.main.ClientReceiverInterface#update(de.bsvrz.dav.daf.main.ResultData[])
	 */
	public void update(ResultData[] results) {
		if(results != null) _dataManager.addScriptRequest(results);
	}

	/**
	 * Änderung der Menge. Meldet den Empfang von Objekten an oder ab.
	 *
	 * @see de.bsvrz.dav.daf.main.config.MutableSetChangeListener#update(de.bsvrz.dav.daf.main.config.MutableSet,de.bsvrz.dav.daf.main.config.SystemObject[],
	 *de.bsvrz.dav.daf.main.config.SystemObject[])
	 */
	public void update(MutableSet set, SystemObject[] addedObjects, SystemObject[] removedObjects) {
		for(SystemObject added : addedObjects) {
			try {
				ConnectionManager.subscribeReceiver(_dav, this, added, _dataDescription, _options, _role);
			}
			catch(FailureException e) {
				e.log(ErrorMessage.MULIPLE_SUBSCRIPTIONS + this + ", " + added);
			}
		}

		for(SystemObject removed : removedObjects) {
			ConnectionManager.unsubscribeReceiver(_dav, this, removed, _dataDescription);
			_dataManager.addScriptRequest(new DeleteScriptRequest(_dav, _dataManager, removed));
		}

		_ready = true;
	}


	/** Meldet den Empfang von Daten ab. Dazu wird zuerst der Empfang von Änderungen an der überwachten Menge abgeschaltet. */
	public void unsubscribe() {

		_mutableSet.removeChangeListener(this);


		for(SystemObject script : (List<SystemObject>)_mutableSet.getElements()) {
			ConnectionManager.unsubscribeReceiver(_dav, this, script, _dataDescription);
		}
	}

	/**
	 * Zeigt ob Anmeldung auf alle Skriptobjekte durchgeführt wurde.
	 *
	 * @return	<code>true</code> wenn alle Anmeldungen durchgeführt wurden. <code>false</code> sonst.
	 */
	public boolean isReady() {
		return _ready;
	}
}
