/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.manager.requests;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.exceptions.NoncriticalException;
import de.bsvrz.pua.prot.manager.ProtocolAnswer;
import de.bsvrz.pua.prot.manager.datamanager.DataManager;
import de.bsvrz.pua.prot.util.ProtocolRequestResult;
import de.bsvrz.sys.funclib.losb.datk.AtgProtocolRequest;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

/**
 * Repräsentiert den Auftrag: Löschen eines gespeicherten Protokolls.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class DeleteProtocolRequest extends ProtocolRequest {

	/** Auftragsparameter */
	private ResultData _resultData;

	/** Auftraggeber */
	private SystemObject _client;

	/** Anfrage-Id */
	private long _requestId;

	/**
	 * Erzeugt den Auftrag
	 *
	 * @param dav         Verbindung zum Datenverteiler
	 * @param resultData  Anfrageparameter. resultData.getData() darf nicht null liefern!
	 * @param dataManager Data Manager
	 * @param client      Auftraggeber
	 * @param requestId   Anfrage-Id
	 */
	public DeleteProtocolRequest(ClientDavInterface dav, ResultData resultData, DataManager dataManager, SystemObject client, long requestId) {
		super(dav, new ResultData[]{resultData}, dataManager);
		this._resultData = resultData;
		this._client = client;
		this._requestId = requestId;
	}

	/**
	 * Löscht ein gespeichertes Protokoll.
	 *
	 * @return Ergebnis der Operation.
	 */
	@Override
    public ProtocolRequestResult process() {
		ProtocolRequestResult result = null;

		Data data = _resultData.getData();

		AtgProtocolRequest request = AtgProtocolRequest.getJavaObject(data);

		//Protokoll-Id aus dem Datenfeld auslesen:
		try {
			Long protocolId = SerializerUtil.deserializeId(request.requestData);
			if(protocolId == null) {
				String msg = ErrorMessage.UNEXPECTED_DATA_TYPE + SerializerUtil.deserialize(request.requestData) + " <-> Long";
				debug.error(msg);
				result = new ProtocolRequestResult(msg);
				ProtocolAnswer.sendError(ProtocolAnswer.sender(dav, _client), _requestId, msg);
			}
			else {
				String msg = delete(protocolId);
				if(msg != null) {
					debug.fine(msg);
					ProtocolAnswer.sendError(ProtocolAnswer.sender(dav, _client), _requestId, msg);
					result = new ProtocolRequestResult(msg);
				}
				else {
					// Quittung senden
					ProtocolAnswer.sendAck(ProtocolAnswer.sender(dav, _client), _requestId);
					result = new ProtocolRequestResult();
				}
			}
		}
		catch(FailureException e) {
			debug.error(e.getMessage());
			result = new ProtocolRequestResult(e.getMessage());
		}
		return result;
	}

	/**
	 * Löscht das gespeicherte Protokoll.
	 *
	 * @param l Protokoll-Id
	 *
	 * @return Fehlermeldung oder <code>null</code> im Erfolgsfall.
	 */
	private String delete(Long l) {
		if(l != null) {
			try {
				dataManager.deleteProtocol(l);
				return null;
			}
			catch(NoncriticalException e) {
				return e.getMessage();
			}
		}
		else {
			return ErrorMessage.CAN_NOT_ACCESS_FILE;
		}
	}
}
