/*
 * Copyright 2010 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.manager.status;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.archive.ArchiveAvailabilityListener;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestManager;
import de.bsvrz.dav.daf.main.config.ConfigurationObject;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.manager.taskmanager.ThreadElement;
import de.bsvrz.pua.prot.manager.taskmanager.ThreadPool;
import de.bsvrz.sys.funclib.debug.Debug;

import java.util.*;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.CopyOnWriteArraySet;

/**
 * Diese Klasse bietet Informationen über laufende Threads und Warteschlangen und bietet damit komfortable Zugriffe auf die Elemente von ThreadPool usw.
 *
 * Die Klasse ist als Singleton realisiert.
 *
 * @author Kappich Systemberatung
 * @version $Revision: 0000 $
 */
public class StatePublisher {

	/** Singleton-Instanz dieser Klasse */
	private static final IdentityHashMap<ClientDavInterface, StatePublisher> INSTANCE_MAP = new IdentityHashMap<>();

	/** Leeres Hilfsarray für toArray()-Aufruf */
	private static final ThreadPool[] EMPTY_ARRAY = new ThreadPool[0];

	/** Liste mit Veröffentlichungs-Modulen, die den Threadstatus über Konsole oder Datenverteiler veröffentlichen */
	private final List<StatePublisherTask> _publishers = new CopyOnWriteArrayList<>();

	/** Map mit überwachten Thread-Pools (Value ist der Name des Threadpools) */
	private final Map<ThreadPool, String> _threadPoolsAndNames = Collections.synchronizedMap(new WeakHashMap<ThreadPool, String>());

	private final Set<ArchiveRequestManager> _archiveRequesters = new CopyOnWriteArraySet<>();
	
	private static final Debug _debug = Debug.getLogger();
	private final ClientDavInterface _dav;

	private ArchiveAvailabilityListener _archiveListener = new ArchiveAvailabilityListener() {
		@Override
		public void archiveAvailabilityChanged(final ArchiveRequestManager archive) {
			new Thread(){
				@Override
				public void run() {
					// Muss in eigenem Thread ausgeführt werden um Sendesteuerungs-Thread nicht zu blockieren
					_archiveAvailable = archive.isArchiveAvailable();
					notifyArchiveAvailable(_archiveAvailable);
				}
			}.start();
		}
	};
	private boolean _archiveAvailable = false;

	private StatePublisher(final ClientDavInterface dav) {
		_debug.fine("Registriere Datenverteilerverbindung");
		_dav = dav;
		refreshArchiveListeners();
		addPublisher(new DavPublisher(dav));
		addPublisher(new ConsolePublisher());
	}

	/**
	 * Gibt den StatePublisher zurück
	 * @return StatePublisher
	 * @param dav DaV
	 */
	public static StatePublisher getInstance(final ClientDavInterface dav) {
		synchronized(INSTANCE_MAP){
			if(INSTANCE_MAP.containsKey(dav)){
				return INSTANCE_MAP.get(dav);
			}
			StatePublisher statePublisher = new StatePublisher(dav);
			INSTANCE_MAP.put(dav, statePublisher);
			return statePublisher;
		}
	}

	private void refreshArchiveListeners() {
		final Set<SystemObject> authorities = new HashSet<>();
		for(ThreadPool threadPool : _threadPoolsAndNames.keySet()) {
			ConfigurationObject configAuth = threadPool.getConfigAuth();
			authorities.add(configAuth);
		}
		for(SystemObject authority : authorities) {
			ArchiveRequestManager archive = _dav.getArchive(authority);
			if(_archiveRequesters.add(archive)){
				_archiveAvailable = archive.isArchiveAvailable();
				archive.addArchiveAvailabilityListener(_archiveListener);
				notifyArchiveAvailable(_archiveAvailable);
			}
		}
	}

	/**
	 * Meldet einen ThreadPool an, dessen Thread-Status in Zukunft durch den StatePublisher publiziert wird
	 * @param threadPool ThreadPool
	 * @param name name des ThreadPools
	 */
	public void registerThreadPool(final ThreadPool threadPool, String name){
		_debug.fine("Registriere Threadpool", name);
		_threadPoolsAndNames.put(threadPool, name);
		refreshArchiveListeners();
	}

	private void addPublisher(final StatePublisherTask publishTask) {
		final Timer timer = new Timer("ThreadStatePublisher", true);
		publishTask.setDataSource(this);
		if(_archiveAvailable) {
			publishTask.notifyArchiveAvailable(_archiveAvailable);
		}
		timer.schedule(publishTask, 0, publishTask.getInterval());
		_publishers.add(publishTask);
	}

	public Status getStatus() {
		List<ThreadPool.QueuedThread> waitingProtocols = new ArrayList<>();
		List<ThreadElement> threadElements = new ArrayList<>();
		Set<ThreadPool> threadPools = _threadPoolsAndNames.keySet();
		// Array-Kopie wegen Threadsicherheit erzeugen,
		// nicht "threadPools.toArray(new ThreadPool[threadPools.length()])"
		// weil die _threadPoolsAndNames-Map zwischen den beiden Aufrufen die Größe ändern könnte!
		// noinspection ToArrayCallWithZeroLengthArrayArgument
		ThreadPool[] arrayCopy = threadPools.toArray(EMPTY_ARRAY);
		for(ThreadPool threadPool : arrayCopy) {
			waitingProtocols.addAll(threadPool.getWaitingProtocols());
			threadElements.addAll(threadPool.getThreads());
		}

		return new Status(waitingProtocols, threadElements);
	}

	/**
	 * Wird aufgerufen, wenn sich ein Thread beendet hat
	 * @param threadElement Thread
	 */	
	public void notifyThreadFinished(final ThreadElement threadElement) {
		for(final StatePublisherTask publisher : _publishers) {
			publisher.notifyThreadFinished(threadElement);
		}
	}

	/**
	 * Wird aufgerufen, wenn ein Protokoll gelöscht wurde
	 * @param protocolId Protokoll-ID
	 */
	public void notifyProtocolRemoved(final long protocolId) {
		for(final StatePublisherTask publisher : _publishers) {
			publisher.notifyProtocolRemoved(protocolId);
		}
	}

	/**
	 * Wird aufgerufen, wenn ein Protokoll gelesen wurde
	 * @param protocolId Protokoll-ID
	 */
	public void notifyProtocolRead(final long protocolId) {
		for(final StatePublisherTask publisher : _publishers) {
			publisher.notifyProtocolRead(protocolId);
		}
	}

	/**
	 * Wird aufgerufen, wenn sich der Zustand des Archivsystems ändert
	 * @param archiveAvailable Ob das Archivsystem verfügbar ist, oder nicht
	 */
	public void notifyArchiveAvailable(final boolean archiveAvailable) {
		for(final StatePublisherTask publisher : _publishers) {
			publisher.notifyArchiveAvailable(archiveAvailable);
		}
	}

	/** Speichert Informationen über Threads, die laufen und sich in der Warteschlange befinden */
	public static class Status {

		private final List<ThreadPool.QueuedThread> _queuedThreads;

		private final List<ThreadElement> _activeThreads;

		/**
		 * Erstellt eine neue Status-Instanz
		 *
		 * @param queuedThreads Threads in Warteschlange
		 * @param activeThreads Aktive Threads
		 */
		private Status(final List<ThreadPool.QueuedThread> queuedThreads, final List<ThreadElement> activeThreads) {
			_queuedThreads = queuedThreads;
			_activeThreads = activeThreads;
		}

		public List<ThreadPool.QueuedThread> getQueuedThreads() {
			return Collections.unmodifiableList(_queuedThreads);
		}

		public List<ThreadElement> getActiveThreads() {
			return Collections.unmodifiableList(_activeThreads);
		}

		@Override
		public String toString() {
			return "Status{" + "_queuedThreads=" + _queuedThreads + ", _activeThreads=" + (_activeThreads == null ? null : Arrays.asList(_activeThreads)) + '}';
		}
	}

	@Override
	public String toString() {
		return "ThreadStatePublisher{" + "_publishers=" + _publishers + '}';
	}
}
