/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing.onlinebuffer;

import de.bsvrz.dav.daf.accessControl.UserInfo;
import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.pua.prot.Protocols;
import de.bsvrz.pua.prot.util.Qualifier;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.exceptions.LoggerException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;
import de.bsvrz.sys.funclib.losb.util.Tuple;

import java.util.*;

/**
 * Klasse zum Empfang von Daten vom Datenverteiler. Es wird eine Anmeldung am Datenverteiler durchgeführt. Mehrere Objekte können sich diese Anmeldung "teilen".
 * Daten werden so lange vorgehalten, bis jeder angemeldete Empfänger die Daten abgeholt hat.
 *
 * @author Martin Hilgers
 */
@SuppressWarnings("deprecation")
class OnlineDataReceiverObject implements ClientReceiverInterface {

	/** Debug Ausgabe */
	private static final Debug _debug = Debug.getLogger();

	/** Legt fest wann die bereits aus dem Eingangspuffer gelesenen Daten gelöscht werden können. */
	private static final int RESIZE_INCOMING = 1000;

	/** DI auf die sich der Receiver anmeldet. */
	private final Qualifier _qualifier;

	/**
	 * Registriertes Objekt -> Index. Objekte an die die Daten ausgeliefert werden sollen. Der Long Wert gibt den Index der Daten an, die ausgeliefert werden
	 * sollen.
	 */
	private final Hashtable<Object, Integer> _subscribers;

	/** Eingehende Daten */
	private final ArrayList<ResultData> incoming;

	/** Zählt die Aufrufe von {@link #take(long,Object)}. */
	private int _takeCalls;

	/** Empfängergruppe, zu der dieser Empfänger gehört. */
	private final OnlineDataReceiverSet _receiverSet;

	/** Objekt das benachrichtigt wird, wenn neue Daten vorliegen. */
	private final NewDataCollector _newDataCollector;
	
	/** Datenverteiler-Verbindung */
	final ClientDavInterface _connection;
	
	/** Benutzer, dessen Rechte geprüft werden sollen */
	final UserInfo _userInfo;

	/**
	 * @param receiverSet        Empfängergruppe, zu der dieser Empfänger gehört.
	 * @param qualifier          Datenidentifikation auf die sich der Receiver anmelden soll.
	 * @param newDataCollector   Objekt das benachrichtigt wird, wenn neue Daten vorliegen.
	 * @param connection         Datenverteilerverbindung
	 * @param userInfo           Benutzer, dessen Rechte geprüft werden sollen
	 */
	public OnlineDataReceiverObject(OnlineDataReceiverSet receiverSet, Qualifier qualifier, NewDataCollector newDataCollector, ClientDavInterface connection, final UserInfo userInfo) {
		this._newDataCollector = newDataCollector;
		_connection = connection;
		_userInfo = userInfo;
		_subscribers = new Hashtable<>();
		incoming = new ArrayList<>();
		_takeCalls = 0;
		this._receiverSet = receiverSet;
		this._qualifier = qualifier;
	}

	/**
	 * Zeigt an ob gerade Daten vorliegen.
	 *
	 * @return {@code true} falls aktuell Daten für diesen Empfänger vorliegen. {@code false} falls bereits alle Daten von allen Empfängern abgeholt
	 *         wurden.
	 */
	public synchronized boolean dataAvailable() {
		for(Integer index : _subscribers.values()) {
			if(index < incoming.size()) return true;
		}
		return false;
	}

	/**
	 * Meldet dieses Objekt als Empfänger für die gegebenen Daten beim DAV an. Fängt alle Ausnahmen ab.
	 *
	 * @param receiverOptions Anmeldeparameter. Normal oder Delta-daten, nachgelieferte Daten.
	 *
	 * @throws FailureException Falls es bei der Kommunikation mit der Konfiguration zu einem Fehler kommt.
	 */
	public void subscribe(ReceiveOptions receiverOptions) throws FailureException {
		DataDescription dataDescription = new DataDescription(_qualifier.attributeGroup, _qualifier.aspect);

		Protocols.checkPermissions(_userInfo, _qualifier.object, dataDescription);

		try {
			_connection.subscribeReceiver(
					this,
					_qualifier.object,
					dataDescription,
					receiverOptions,
					ReceiverRole.receiver()
			);
		}
		catch(Exception e) {
			throw new FailureException(ErrorMessage.COMMUNICATION, e, LoggerException.WARNING);
		}
	}

	/**
	 * Empfängt die Daten vom Datenverteiler. Datensätze, deren Zeitstempel nicht im Zeitbereich des Protokolls liegt, werden ignoriert.
	 *
	 * @see ClientReceiverInterface#update(ResultData[])
	 */
	public void update(ResultData[] results) {
		long timeStamp;
		for(ResultData result : results) {
			if(result.hasData()) {
				// OPT optimieren...
				timeStamp = result.getDataTime();
				for(int i = 0; i < _receiverSet.getPeriods().size(); i++) {
					Tuple<Long, Long> period = _receiverSet.getPeriods().get(i);
					if(period.first <= timeStamp && timeStamp <= period.last) {
						storeData(result);
						_newDataCollector.newDataAvailable(); // Benachrichtigt Empfänger, dass neue Daten vorliegen.
						break;
					}
				}
			}
		}
	}

	/**
	 * Speichert den eingegangenen Datensatz.
	 *
	 * @param result Eingegangener Datensatz
	 */
	protected void storeData(ResultData result) {
		synchronized(this) {
			incoming.add(result);
			notifyAll();
		}
	}

	/**
	 * Liefert einen Datensatz. Methode wartet bis ein Datensatz eingetroffen ist.
	 *
	 * @param subscriber Objekt für das die Daten bestimmt sind. Darf nicht {@code null} sein.
	 * @param timeout    Maximale Wartezeit. Wird die Wartezeit überschritten, so wird die Methode abgebrochen und {@code false} zurückgegeben. Ist dieser
	 *                   Parameter {@code 0} so wird solange gewartet, bis Daten eintreffen.
	 *
	 * @return Empfangener Datensatz. {@code null} falls {@code subscriber} kein angemeldeter Empfänger.
	 *
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	public synchronized ExtResultData take(long timeout, Object subscriber) throws InterruptedException {
		Integer index = _subscribers.get(subscriber);
		if(index != null) {
			if(index >= incoming.size()) // hier MUSS if stehen, und nicht while! Sonst gerät man in eine
			// Endlosschleife, wenn keine Daten eingehen!
			{
				wait(timeout);
			}
			if(index < incoming.size()) {
				index = _subscribers.get(subscriber);
			}
			else {
				return null;
			}

			ResultData resultData = incoming.get(index);
			// OPT besser gleich beim Eingang der Daten durchführen. Dann muss das nur einmal gemacht werden
			ExtResultData result = new ExtResultData(resultData, resultData.getData(), resultData.getDataIndex());
			_subscribers.put(subscriber, ++index); // NICHT index++!

			// ab un an Daten alte Daten aus incoming löschen
			_takeCalls++;
			if(incoming.size() > RESIZE_INCOMING && _takeCalls > RESIZE_INCOMING) {
				_takeCalls = 0;
				int deliveredTill = incoming.size();
				for(Integer i : _subscribers.values()) {
					if(i < deliveredTill) deliveredTill = i;
				}
				if(deliveredTill > 0) {
					incoming.subList(0, deliveredTill).clear();
				}
				for(final Map.Entry<Object, Integer> entry : _subscribers.entrySet()) {
					_subscribers.put(entry.getKey(), entry.getValue() - deliveredTill);
				}
			}

			return result;
		}
		else {
			return null;
		}
	}

	/**
	 * Meldet den Empfang von Daten ab. Das Objekt meldet die Datenidentifikation erst vom Empfang der Daten ab, wenn es keine Interessenten mehr dafür gibt.
	 *
	 * @param object Empfänger der abgemeldet werden soll.
	 */
	public synchronized void unsubscribe(Object object) {
		_subscribers.remove(object);
		if(_subscribers.isEmpty()) {
			try {
				_connection.unsubscribeReceiver(
						this, _qualifier.object, new DataDescription(_qualifier.attributeGroup, _qualifier.aspect)
				);
			}
			catch(ConfigurationException e) {
				_debug.error(ErrorMessage.COMMUNICATION, e);
			}
		}
	}

	/**
	 * Registriert eine weitere Anmeldung auf die angemeldeten Daten.
	 *
	 * @return Objekt zum Zugriff auf die eingehenden Daten.
	 */
	public synchronized OnlineDataReceiver registerNewSubscriber() {
		OnlineDataReceiver odr = new OnlineDataReceiver(this);
		_subscribers.put(odr, 0);
		return odr;
	}
}
