/*
 * Copyright 2014 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.util;

import de.bsvrz.dav.daf.main.config.*;
import de.bsvrz.pua.prot.manager.ProtocolRelatedOperation;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.SerializerUtil;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;

import java.io.Serializable;

/**
 * Diese Klasse bildet einen in Bearbeitung befindlichen Auftrag ab
 *
 * @see de.bsvrz.pua.prot.client.PuaClient#getJobList(ProtocolRequestResult)
 *
 * @author Kappich Systemberatung
 * @version $Revision$
 */
public class JobInProgress implements Serializable {
	private static final long serialVersionUID = 8847657996087883139L;

	/**
	 * Datenmodell zur Auflösung des Operationscodes
	 */
	private transient DataModel _dataModel;

	/**
	 * Die Auftrags-Id dieses Auftrags
	 */
	private long _jobId;

	/**
	 * Anfragekennung/Operationscode des protokollbezogenen Auftrags. Dies entspricht den Statuswerten in att.puaProtokollOperationsCode.
	 *
	 * Mögliche Werte sind:
	 * <ul>
	 *     <li>{@link de.bsvrz.pua.prot.manager.ProtocolRelatedOperation#CREATE} (Protokoll erstellen)
	 *     <li>{@link de.bsvrz.pua.prot.manager.ProtocolRelatedOperation#SEND_SAVED_PROTOCOL} (Gespeichertes Protokoll abrufen)
	 *     <li>{@link de.bsvrz.pua.prot.manager.ProtocolRelatedOperation#SEND_UNREAD_PROTOCOL} (Ungelesenes Protokoll abrufen)
	 * </ul>
	 *
	 * Clients sollten sich darauf einstellen, dass in Zukunft evtl. weitere Operationscodes hinzugefügt werden.
	 *
	 * @see de.bsvrz.pua.prot.manager.ProtocolRelatedOperation
	 */
	private int _operationCode;

	/**
	 * True wenn in Bearbeitung
	 */
	private JobState _jobState;

	/**
	 * Aktueller Bearbeitungsfortschritt (0-100)
	 */
	private byte _status;

	/**
	 * Startzeit wie in {@link System#currentTimeMillis()}
	 */
	private long _startTime;

	/**
	 * ID des Client-System-Objekts
	 */
	private long _clientId;

	/**
	 * Serialisierte Parameter des Auftrags
	 */
	private byte[] _parameters = new byte[0];

	private static final Debug _debug = Debug.getLogger();

	/**
	 * Erstellt einen neuen JobInProgress
	 * @param jobId Auftrags-ID
	 * @param operationCode Operationscode
	 * @param jobState Flag "in Bearbeitung"
	 * @param status Bearbeitungsfortschritt
	 * @param startTime Startzeit
	 * @param clientId Client-ID
	 * @param parameters Parameter des Auftrags (Operationsabhängig)
	 * @throws de.bsvrz.sys.funclib.losb.exceptions.FailureException Exception
	 */
	public JobInProgress(final long jobId, final int operationCode, final JobState jobState, final byte status, final long startTime, final long clientId, final Object parameters) throws FailureException {
		_jobId = jobId;
		_operationCode = operationCode;
		_jobState = jobState;
		_status = status;
		_startTime = startTime;
		_clientId = clientId;
		if(parameters instanceof ProcessingParameter) {
			ProcessingParameter processingParameter = (ProcessingParameter) parameters;
			_parameters = processingParameter.serialize();
		}
		else if(parameters instanceof Serializable) {
			_parameters = SerializerUtil.serializeToByteArray((Serializable) parameters);
		}
	}

	/**
	 * Interne Operation zum setzen des Datenmodells
	 * @param dataModel Datenmodell
	 */
	public void initialize(final DataModel dataModel) {
		_dataModel = dataModel;
	}

	/**
	 * Gibt die Anfrage-ID des Auftrags zurück. Mit dieser Id kann der Auftrag z.B. angehalten oder abgebrochen werden.
	 *
	 * @return Auftrags-ID.
	 */
	public long getJobId() {
		return _jobId;
	}

	/**
	 * Anfragekennung/Operationscode des protokollbezogenen Auftrags. Dies entspricht den Statuswerten in att.puaProtokollOperationsCode.
	 *
	 * Mögliche Werte sind:
	 * <ul>
	 *     <li>{@link de.bsvrz.pua.prot.manager.ProtocolRelatedOperation#CREATE} (Protokoll erstellen)
	 *     <li>{@link de.bsvrz.pua.prot.manager.ProtocolRelatedOperation#SEND_SAVED_PROTOCOL} (Gespeichertes Protokoll abrufen)
	 *     <li>{@link de.bsvrz.pua.prot.manager.ProtocolRelatedOperation#SEND_UNREAD_PROTOCOL} (Ungelesenes Protokoll abrufen)
	 * </ul>
	 *
	 * Clients sollten sich jedoch darauf einstellen, dass in Zukunft möglicherweise weitere Operationscodes hinzugefügt werden. (Also beispielsweise
	 * Aufträge mit unbekannten Operationscodes ignorieren oder {@link #getOperationName()} benutzen um bei unbekannten Codes zumindest
	 * den Typ des Auftrags anzeigen zu können.)
	 *
	 * @see de.bsvrz.pua.prot.manager.ProtocolRelatedOperation
	 *
	 * @return Operationscode
	 */
	public int getOperationCode() {
		return _operationCode;
	}

	/**
	 * Gibt den Namen der Operation zurück, die sich hinter dem Operationscode verbirgt.
	 * Dieser Name kann z. B. für die Darstellung in einer GUI benutzt werden.
	 * @return Name der Operation oder "" (Leerstring) falls der Name nicht ermittelbar ist.
	 */
	public String getOperationName() {
		if(_dataModel == null) return "";
		IntegerAttributeType type = (IntegerAttributeType) _dataModel.getAttributeType("att.puaProtokollOperationsCode");
		if(type == null) return "";
		for(IntegerValueState state : type.getStates()) {
			if(state.getValue() == _operationCode){
				return state.getName();
			}
		}
		return "";
	}

	/**
	 * Gibt den Berabeitungszustands des Auftrag zurück (z.B: in Warteschlange, in Bearbeitung, beendet)
	 * @return Bearbeitungszustand
	 */
	public JobState getJobState() {
		return _jobState;
	}

	/**
	 * Gibt den Fortschritt des Auftrags zurück.
	 * @return Prozent-Wert von 0-100. Bei Aufträgen in Warteschlange 0.
	 */
	public byte getStatus() {
		return _status;
	}

	/**
	 * Gibt die Startzeit des Auftrags zurück (wann der Auftrag von PuA entgegengenommen wurde)
	 * @return Startzeit in Millisekunden (analog zu {@link System#currentTimeMillis()},
	 * oder -1 falls unbekannt (z. B. aufgrund einer veralteten PuA-Server-Implementierung)
	 */
	public long getStartTime() {
		return _startTime;
	}

	/**
	 * Gibt die Applikations-ID des beauftragenden Clients zurück
	 * @return die Applikations-ID des beauftragenden Clients,
	 * oder -1 falls unbekannt (z. B. aufgrund einer veralteten PuA-Server-Implementierung)
	 */
	public long getClientId() {
		return _clientId;
	}

	/**
	 * Gibt das Applikationsobjekt zurück, das den Auftrag veranlasst hat
	 * @return Applikationsobjekt oder null falls nicht ermittelbar (z. B. aufgrund einer veralteten PuA-Server-Implementierung)
	 */
	public ClientApplication getClient() {
		if(_dataModel == null || _clientId == -1) return null;
		SystemObject object = _dataModel.getObject(_clientId);
		if(object instanceof ClientApplication) {
			return (ClientApplication) object;
		}
		return null;
	}

	/**
	 * Gibt die Parameter des Auftrags zurück.
	 *
	 * <ul>
	 *     <li>Bei Protokoll-Erstellen-Aufträgen ein Objekt vom Typ {@link de.bsvrz.pua.prot.util.ProcessingParameter}.
	 *     <li>Beim Abrufen gespeicherter oder ungelesener Protokolle ein Long mit der Protokoll-Id.
	 * </ul>
	 *
	 * @return Operationsspezifische Parameter des Auftrags oder null falls mit der aktuellen Softwareversion nicht interpretierbar/nicht vorhanden
	 */
	public Object getParameters(){
		if(_operationCode == ProtocolRelatedOperation.CREATE){
			try {
				return ProcessingParameter.deserialize(_dataModel, _parameters);
			}
			catch(FailureException e) {
				_debug.warning("Kann Parameter nicht deserialisieren", e);
				return null;
			}
		}
		return SerializerUtil.deserialize(_parameters);
	}

	@Override
	public String toString() {
		String operationName = getOperationName();
		if(operationName.equals("")) operationName = "Unbekannt";
		return "Auftrag " + _jobId + " (" + operationName + ") " + _status + "%";
	}
}
