/*
 * Copyright 2018 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.util;

import com.google.common.collect.ImmutableList;

import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.util.*;

/**
 * Diese Klasse kapselt einen nicht notwendigerweise zusammenhängenden Zeitbereich für PuA-Anfragen.
 * <p> 
 * Diese Klasse ist prinzipiell eine Collection von {@link DiscretePeriod}-Objekten.
 * <p>
 * Diese Klasse ist unveränderlich (immutable) und daher threadsafe.
 * 
 * @author Kappich Systemberatung
 */
public class PeriodSet {

	/**
	 * Liste mit Zeitbereichen
	 */
	private final ImmutableList<DiscretePeriod> _periods;

	/** 
	 * Erstellt ein neues PeriodSet (privater Konstruktor), of(...)-Methoden verwenden um Instanz zu erhalten.
	 * @param periods Enthaltene Zeitbereiche
	 * @see #of()   
	 */
	private PeriodSet(final ImmutableList<DiscretePeriod> periods) {
		_periods = periods;
	}

	/** 
	 * Gibt die enthaltenen Zeitbereiche zurück
	 * @return die enthaltenen Zeitbereiche
	 */
	public ImmutableList<DiscretePeriod> getPeriods() {
		return _periods;
	}

	/**
	 * Sortiert die enthaltenen Zeitbereiche und fasst dabei überlappende Zeitbereiche zusammen (Vereinigungsmenge)
	 * @return Sortierte Kopie dieses Objekts. Das {@code this}-Objekt bleibt unverändert.
	 */
	public PeriodSet order() {
		final Deque<DiscretePeriod> periods = new ArrayDeque<>();
		_periods.stream().sorted(Comparator.comparing(DiscretePeriod::getFrom)).forEach(it -> {
			if(periods.isEmpty() || periods.getLast().getTo().isBefore(it.getFrom())){
				periods.addLast(it);
			}
			else if(it.getTo().isAfter(periods.getLast().getTo())) {
				periods.addLast(DiscretePeriod.of(periods.removeLast().getFrom(), it.getTo()));
			}
		});
		return copyOf(ImmutableList.copyOf(periods));
	}

	/**
	 * Erstellt ein einfaches PeriodSet basierend auf Millisekunden-Zeitstempeln seit Epoch
	 * @param fromMillis von
	 * @param toMillis   bis
	 * @return DiscretePeriod
	 */
	public static PeriodSet of(long fromMillis, long toMillis) {
		return of(DiscretePeriod.of(fromMillis, toMillis));
	}

	/**
	 * Erstellt ein einfaches PeriodSet basierend auf {@link LocalDateTime}-Objekten in der Standard-Zeitzone.
	 * @param from von
	 * @param to   bis
	 * @return DiscretePeriod
	 */
	public static PeriodSet of(LocalDateTime from, LocalDateTime to) {
		return of(DiscretePeriod.of(from, to));
	}

	/**
	 * Erstellt ein einfaches PeriodSet basierend auf {@link LocalDateTime}-Objekten in der angegebenen Zeitzone.
	 * @param from von
	 * @param to   bis
	 * @param zone Zeitzone
	 * @return DiscretePeriod
	 */
	public static PeriodSet of(LocalDateTime from, LocalDateTime to, final ZoneId zone) {
		return of(DiscretePeriod.of(from, to, zone));
	}

	/**
	 * Erstellt ein einfaches PeriodSet basierend auf {@link Instant}-Objekten.
	 * @param from von
	 * @param to   bis
	 * @return DiscretePeriod
	 */
	public static PeriodSet of(Instant from, Instant to) {
		return of(DiscretePeriod.of(from, to));
	}
	
	/**
	 * Erstellt ein PeriodSet aus einem oder mehreren {@link DiscretePeriod}-Objekten
	 * @param first  Erstes Element (nicht null)
	 * @param others Weitere Elemente (nicht null)
	 * @return DiscretePeriod
	 */
	public static PeriodSet of(DiscretePeriod first, DiscretePeriod... others) {
		ImmutableList.Builder<DiscretePeriod> builder = ImmutableList.builder();
		builder.add(first);
		builder.addAll(Arrays.asList(others));
		return new PeriodSet(builder.build());
	}

	/**
	 * Erstellt ein leeres PeriodSet
	 * @return DiscretePeriod
	 */
	public static PeriodSet of() {
		return new PeriodSet(ImmutableList.of());
	}
	
	/**
	 * Erstellt ein PeriodSet aus mehreren {@link DiscretePeriod}-Objekten
	 * @param periods Elemente (darf nicht null enthalten)
	 * @return DiscretePeriod
	 */
	public static PeriodSet copyOf(final Collection<DiscretePeriod> periods) {
		return new PeriodSet(ImmutableList.copyOf(periods));
	}

	@Override
	public boolean equals(final Object o) {
		if(this == o) return true;
		if(!(o instanceof PeriodSet)) return false;

		final PeriodSet periodSet = (PeriodSet) o;

		return _periods.equals(periodSet._periods);

	}

	@Override
	public int hashCode() {
		return _periods.hashCode();
	}

	@Override
	public String toString() {
		return _periods.toString();
	}
}
