/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.client.dataobject;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.Attribute;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeType;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.data.ArrayProtocolData;
import de.bsvrz.pua.prot.data.AttributeGroupProtocolData;
import de.bsvrz.pua.prot.data.AttributeProtocolData;
import de.bsvrz.pua.prot.data.MultiColumnProtocolData;
import de.bsvrz.pua.prot.data.ProtocolData;
import de.bsvrz.pua.prot.util.ProcessingInformation;
import de.bsvrz.pua.prot.util.ResultValue;
import de.bsvrz.sys.funclib.losb.util.Util;

import java.util.*;


/**
 * Fasst die Daten eines Aggregations-Ergebnisdatensatzes zusammen
 *
 * @version $Revision$ / $Date$
 */
public class ProtocolAggregationResultData implements ProtocolResult {

	private static final long serialVersionUID = -8230861341084136611L;

	/** Ergebnisdatensatz gehört zu der gesetzten Aggregation. */
	public ProcessingInformation.ApplyAggregation order;

	/** Werte */
	public ResultValue values[];

	/** Liste der Startzeitstempel */
	public long[] startTimeStamps;

	/** Liste der Endzeitstempel */
	public long[] endTimeStamps;

	/** Status */
	public byte status;

	private final int _intervalIndex;

	private ProtocolAdministrationData _adminData;

	private ProtocolAggregationData _aggregationData;


	/** @see Object#toString() */
	@Override
    public String toString() {
		StringBuffer sb = new StringBuffer();
		//nicht aufgefüllten Zeitstempel ausgeben (-> ist immer der größte Zeitstempel)

		sb.append("ADS ");
		Util.msToDate(sb, getStartTimeStamp());
		sb.append("-");
		Util.msToDate(sb, getEndTimeStamp());
		sb.append(":\t");
		for(ResultValue value : values) {
			sb.append(value).append("\t");
		}
		sb.append("Status: ").append(status).append("%\t");
		sb.append(order);
		return sb.toString();
	}


	/**
	 * Liefert den End-Zeitstempel zurück. Es wird der größte End-Zeitstempel zurückgegeben.
	 *
	 * @return Zeitstempel.
	 */
	public long getEndTimeStamp() {
		long maxTime = Long.MIN_VALUE;
		for(long time : endTimeStamps) {
			if(maxTime < time && time > 0) maxTime = time;
		}
		return maxTime;
	}

	/**
	 * Liefert den Start-Zeitstempel zurück. Es wird der kleinste Start-Zeitstempel zurückgegeben.
	 *
	 * @return Zeitstempel.
	 */
	public long getStartTimeStamp() {
		long minTime = Long.MAX_VALUE;
		for(long time : startTimeStamps) {
			if(minTime > time && time > 0) minTime = time;
		}
		return minTime;
	}


	/**
	 * @param values            Werte.
	 * @param order             Anwendung der Aggregationen.
	 * @param status            Status
	 * @param endTimeStamps EndTimeStamps
	 * @param intervalIndex IntervalIndex
	 * @param startTimeStamps StartTimeStamps
	 */
	public ProtocolAggregationResultData(
			ResultValue values[], final long[] startTimeStamps, final long[] endTimeStamps, ProcessingInformation.ApplyAggregation order, byte status, int intervalIndex) {
		this.values = values;
		this.startTimeStamps = startTimeStamps;
		this.endTimeStamps = endTimeStamps;
		this.order = order;
		this.status = status;
		_intervalIndex = intervalIndex;
	}

	/** @see Object#equals(Object) */
	@Override
    public boolean equals(Object object) {
		if(object == null) {
			return false;
		}
		else if(object instanceof ProtocolAggregationResultData) {
			ProtocolAggregationResultData other = (ProtocolAggregationResultData)object;

			return Arrays.equals(this.startTimeStamps, other.startTimeStamps) && Arrays.equals(this.endTimeStamps, other.endTimeStamps)
			       && this.status == other.status && Arrays.equals(this.values, other.values) && Objects.equals(this.order, other.order);
		}
		else {
			return false;
		}
	}

	@Override
	public int hashCode() {
		return (int)((getStartTimeStamp() & 0xffff) + (getEndTimeStamp() & 0x0f0f) * 47);
	}

	/** @see ProtocolResult#getType() */
	public int getType() {
		return ProtocolResult.AGGREGATION_RESULT_DATA;
	}

	public ProcessingInformation.ApplyAggregation getOrder() {
		return order;
	}

	@Override
	public String toString(final DataModel dataModel) {
		StringBuilder sb = new StringBuilder();
		//nicht aufgefüllten Zeitstempel ausgeben (-> ist immer der größte Zeitstempel)

		sb.append(String.format("%-28s\n", getIntervalIndex() + " "  + order));

		LinkedHashSet<String> aggregations = new LinkedHashSet<>();
		for(int i = 0; i < getColumnCount(); i++){
			aggregations.addAll(getAggregations(i));
		}
		for(Iterator<String> iterator = aggregations.iterator(); iterator.hasNext(); ) {
			final String aggregation = iterator.next();
			sb.append(String.format("%-28s\t", aggregation));
			for(int i = 0; i < getColumnCount(); i++) {
				ProtocolData data = getData(dataModel, i, aggregation);
				sb.append(String.format("%28s\t", data == null ? "" :data.valueToString()));
			}
            if(iterator.hasNext()){
	            sb.append("\n");
            }
        }
		return sb.toString();
	}

	/**
	 * Wird intern aufgerufen, um clientseitig der Zeile den Spaltenkopf zu übergeben
	 * @param adminData Spaltenkopf
	 */
	public void setAdminData(final ProtocolAdministrationData adminData) {
		_adminData = adminData;
	}

	/**
	 * Gibt eine Liste der Data-Objekte zurück
	 * @param dataModel DataModel
	 * @return Liste mit Data-Objekten
	 */
	public List<ProtocolData> getDataList(final DataModel dataModel){
		final List<ProtocolData> result = new ArrayList<>(values.length);
		for(int i = 0; i < values.length; i++) {
			result.add(getData(dataModel, i) );
		}
		return result;
	}

	/**
	 * Gibt ein Data-Objekt zurück, dass die ganze Zeile beinhaltet
	 * @param dataModel DataModel
	 * @return Data-Objekt
	 */
	public ProtocolData getData(final DataModel dataModel) {
		final List<ProtocolData> list = getDataList(dataModel);
		return new MultiColumnProtocolData(list, _adminData.columns);
	}

	/**
	 * Gibt ein Data-Objekt (als Array) für alle Aggregationen zurück.
	 * @param dataModel Datenmodell
	 * @param index Spaltenindex
	 * @return Data-Objekt
	 */
	public ProtocolData getData(final DataModel dataModel,final int index){
		final ResultValue value = values[index];
		final Column column = _adminData.columns[index];
		return createDataArray(dataModel, value, column);
	}

	/**
	 * Gibt ein Data-Objekt für eine bestimmte Aggregation zurück
	 * @param dataModel Datenmodell
	 * @param index Spaltenindex
	 * @param aggregationIndex Index der Aggregation, dessen Wert zurückgegeben werden soll. Der Index entspricht dem Index der Aggregation von
	 * {@link #getAggregations(int)}
	 * @return Data-Objekt
	 */
	public ProtocolData getData(final DataModel dataModel, final int index, final int aggregationIndex){
		final ResultValue value = values[index];
		final Column column = _adminData.columns[index];
		if(value.getChildren().size() <= aggregationIndex){
			return null;
		}
		return createDataObject(dataModel, value.getChildren().get(aggregationIndex), column);
	}

	/**
	 * Gibt ein Data-Objekt für eine bestimmte Aggregation zurück.
	 * @param dataModel Datenmodell
	 * @param index Spaltenindex
	 * @param aggregation Aggregation, dessen Wert zurückgegeben werden soll. Die vorhandenen Aggregationen können mit {@link #getAggregations(int)} ausgelesen
	 * werden.
	 * @return Data-Objekt oder null wenn die angegebene Aggregation für diese Spalte bzw. Zelle nicht definiert ist.
	 */
	public ProtocolData getData(final DataModel dataModel,final int index, final String aggregation){
		final int aggregationIndex = getAggregations(index).indexOf(aggregation);
		if(aggregationIndex == -1){
			return null;
		}
		return getData(dataModel, index, aggregationIndex);
	}

	/**
	 * Gibt die definierten Aggregationen für den angegebenen Spaltenindex zurück
	 *
	 * @param index Spaltenindex
	 * @return Liste mit den Aggregationsnamen, die für diese Zelle definiert sind
	 */
	public List<String> getAggregations(final int index){
		return _aggregationData.getAggregations().get(index);
	}

	/**
	 * Gibt die Aggregationen für eine Zelle mit enthaltenen Data-Objekten verknüpft zurück
	 * @param dataModel Datenmodell
	 * @param index Spaltenindex
	 * @return Eine Map Aggregationsname-&gt;Wert
	 */
	public Map<String, ProtocolData> getAggregations(final DataModel dataModel, final int index){
		final List<String> aggregations = getAggregations(index);
		final Map<String, ProtocolData> map = new HashMap<>(aggregations.size());
		for(int i = 0, aggregationsSize = aggregations.size(); i < aggregationsSize; i++) {
			final String aggregationName = aggregations.get(i);
			map.put(aggregationName, getData(dataModel, index, i));
		}
		return map;
	}

	/**
	 * Gibt die Aggregationen für eine ganze Zeile mit enthaltenen Data-Objekten verknüpft zurück
	 * @param dataModel Datenmodell
	 * @return Eine Liste über die Spaltenindizes, deren Einträge Maps enthalten, welche die Aggregationen für eine Zelle enthalten und den Data-Objekten zuordnen
	 */
	public List<Map<String, ProtocolData>> getAggregations(final DataModel dataModel){
		final List<Map<String, ProtocolData>> result = new ArrayList<>();
		for (int i = 0; i < values.length; i++){
			result.add(getAggregations(dataModel, i) );
		}
		return result;
	}

	/**
	 * Erstellt ein Data-Objekt
	 * @param dataModel Datenmodell
	 * @param value Wert des Datums
	 * @param column Spalte des Datums um Attributtyp usw. zu bestimmen
	 * @return Data-Objekt
	 */
	static ProtocolData createDataObject(final DataModel dataModel, final ResultValue value, final Column column) {
		final long referenceId = column.getReference();
		if(column.getType() == Column.ColumnType.DURATION_ATTRIBUTE){
			// Bei Zeitdauer-Attributen wird keine Attribut-Referenz mitgeliefert, es erscheint sinnvoll, hier einen passenden Attributtyp manuell zu wählen
			return new AttributeProtocolData(dataModel.getAttributeType("att.zeitDauer"), dataModel, value);
		}
		final SystemObject systemObject = dataModel.getObject(referenceId);
		if(systemObject instanceof Attribute) {
			final Attribute attribute = (Attribute)systemObject;
			return new AttributeProtocolData(attribute, dataModel, value);
		}
		if(systemObject instanceof AttributeType) {
			final AttributeType attributeType = (AttributeType)systemObject;
			return new AttributeProtocolData(attributeType, dataModel, value);
		}
		if(systemObject instanceof AttributeGroup) {
			final AttributeGroup attributeGroup = (AttributeGroup)systemObject;
			return new AttributeGroupProtocolData(attributeGroup, dataModel, value);
		}
		throw new IllegalArgumentException("Ungültige Attribut-ID im Spaltenkopf: " + referenceId + " (Systemobject: " + systemObject + ")");
	}

	private static ProtocolData createDataArray(final DataModel dataModel, final ResultValue value, final Column column) {
		final ArrayList<Data> dataList = new ArrayList<>();
		for(final ResultValue resultValue : value.getChildren()) {
			final ProtocolData data = createDataObject(dataModel, resultValue, column);
			dataList.add(data);
		}
		return new ArrayProtocolData(dataList);
	}

	/**
	 * Gibt den Zeitbereichs-Index zurück
	 * @return Der Index der angefragten Zeitbereiche
	 */
	public int getIntervalIndex() {
		return _intervalIndex;
	}

	/**
	 * Interne Funktion um dem ProtocolAggregationResultData ein ProtocolAggregationData zuzuordnen
	 * @param aggregationData ProtocolAggregationData
	 */
	public void setAggregationData(final ProtocolAggregationData aggregationData) {
		_aggregationData = aggregationData;
	}

	/**
	 * Gibt die Anzahl Spalten zurück
	 * @return die Anzahl Spalten
	 */
	public int getColumnCount(){
		return _adminData.getColumnCount();
	}
}
