/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.interpreter.semantics;

import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.SemanticErrorException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.util.*;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.datk.NoChangeMarker;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.util.*;

/**
 * Diese Klasse bietet Methoden zum Überprüfen der Semantik eines Skripts. Semantics wird vom {@link de.bsvrz.pua.prot.interpreter.ProtocolParser} und {@link
 * de.bsvrz.pua.prot.interpreter.ProtocolTreeWalker} verwendet, um die Semantik des Skripts zu überprüfen. Falls {@link #processScript} true ist, dann wird
 * überprüft, ob jede Qualifizierung ein Objekt enthält. Enthält sie statt des Objekts einen Objekttyp, so wird eine {@link CriticalParserException} geworfen.
 * Dies ist notwendig, da zur Datenaufbereitung konkrete Objekte benötigt werden. Vor der Verwendung von {@link #pi} durch andere Klassen muss {@link
 * de.bsvrz.pua.prot.interpreter.semantics.SemanticChecks#cleanUp(ProcessingParameter,boolean)} aufgerufen werden, um abschließende semantische Überprüfungen
 * durchzuführen.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Semantics implements SemanticChecks {

	/** Debug Ausgabe */
	private static final Debug _debug = Debug.getLogger();

	/** Verwendetes DatenModel */
	protected DataModel model;

	/** Objekt das Typüberprüfungen durchführt */
	private MemberCheck check;

	/** Symboltabelle */
	protected SymbolTable symbolTable;

	/** Datenobjekt, in dem alle zur Datenaufbereitung notwendigen Daten abgelegt werden */
	protected ProcessingInformation pi;

	/**
	 * True falls das Skript vollständig ausgewertet werden soll, d.h. wenn später die Datenaufbereitung auf pi gestartet werden soll.<br> Bei 'False' wird zum
	 * Beispiel nicht darauf bestanden, dass eine Bindung für jedes Objekt existiert
	 */
	protected boolean processScript;

	/** Liste mit allen Objekten, die Semantiküberprüfungen durchführen. */
	private List<SemanticChecks> _semanticChecks;

	/** Objekt das die Semantik von Aggregationen überprüft */
	private AggregationSemantics _aggregationSemantics;

	/** Objekt das die Semantik von Attributgruppen überprüft */
	private AttributeGroupSemantics _attributeGroupSemantics;

	/** Objekt das die Semantik von Attributen aus dem Spaltenbereich überprüft */
	private AttributeSemantics _attributeSemantics;

	/** Objekt das die Semantik von Ausdrücken überprüft */
	private ExpressionSemantics _expressionSemantics;

	/** Objekt das die Semantik von diversen Konstrukten überprüft */
	private MiscellaneousSemantics _miscSemantics;

	/** Objekt das die Semantik von Realen Attributen überprüft */
	private RealAttributeSemantics _realAttributeSemantics;

	/** Objekt das die Semantik des Mit-Konstrukts überprüft */
	private ScopeSemantics _scopeSemantics;

	/** Objekt das die Semantik von temporären Attributen überprüft */
	private TempAttributeSemantics _tempAttributeSemantics;


	/**
	 * Konstruktor. Erzeugt die Objekte zur Semantiküberprüfung.
	 *
	 * @param model  Verbindung zur Konfiguration
	 * @param symbol Symboltabelle
	 * @param pi     Objekt, in das die gewonnenen Informationen abgelegt werden sollen
	 *
	 * @throws CriticalParserException Falls einer der Parameter null ist
	 */
	public Semantics(final DataModel model, final SymbolTable symbol, final ProcessingInformation pi) throws CriticalParserException {
		if(model == null) {
			_debug.error(InterpreterErrorMessage.NO_MODEL);
			throw new SemanticErrorException(InterpreterErrorMessage.NO_MODEL);
		}
		else if(symbol == null) {
			throw new SemanticErrorException(InterpreterErrorMessage.NO_SYMBOL_TABLE);
		}
		else if(pi == null) throw new SemanticErrorException(InterpreterErrorMessage.NO_RETURN_OBJECT_PROVIDED);

		this.model = model;
		this.symbolTable = symbol;
		this.pi = pi;
		//simulationsvariante muss schon hier gesetzt werden, da die anderen Semantics-Klassen darauf zugreifen.
		pi.setSimVar(symbolTable._simVar);

		check = new MemberCheck(model, symbolTable);

		processScript = symbolTable._pp != null;

		_scopeSemantics = new ScopeSemantics(this);
		_attributeSemantics = new AttributeSemantics();
		_realAttributeSemantics = new RealAttributeSemantics(this, _attributeSemantics, _scopeSemantics);
		_aggregationSemantics = new AggregationSemantics(this, symbolTable, _attributeSemantics, _realAttributeSemantics);
		_attributeGroupSemantics = new AttributeGroupSemantics(this, _scopeSemantics, _attributeSemantics, _realAttributeSemantics);
		_expressionSemantics = new ExpressionSemantics(this, _attributeSemantics, _realAttributeSemantics);
		_miscSemantics = new MiscellaneousSemantics(this);
		_tempAttributeSemantics = new TempAttributeSemantics(this, _attributeSemantics);

		_semanticChecks = new ArrayList<>();
		_semanticChecks.add(_scopeSemantics);
		_semanticChecks.add(_attributeSemantics);
		_semanticChecks.add(_realAttributeSemantics);
		_semanticChecks.add(_aggregationSemantics);
		_semanticChecks.add(_attributeGroupSemantics);
		_semanticChecks.add(_expressionSemantics);
		_semanticChecks.add(_miscSemantics);
		_semanticChecks.add(_tempAttributeSemantics);
	}

	/**
	 * Führt abschließende Semantiküberprüfungen durch und stellt die zur Datenaufbereitung notwendigen Informationen in {@link #pi} zusammen. Dabei werden die
	 * {@link SemanticChecks#cleanUp(ProcessingParameter,boolean)} Methoden aller Objekte, die Semantiküberprüfungen durchführen, aufgerufen. Außerdem wird, falls
	 * {@link #processScript} true ist, überprüft, ob alle Aspekte gebunden werden konnten.
	 *
	 * @param pp            Auftragsparameter
	 * @param makeProcessScript <code>true</code> wenn ein Protokoll erstellt werden soll. <code>false</code> sonst.
	 *
	 * @throws SemanticErrorException Falls es zu einem Fehler bei der Überprüfung kommt
	 */
	public void cleanUp(final ProcessingParameter pp, final boolean makeProcessScript) throws SemanticErrorException {
		/*
		 * Es werden alle Einträge aus der Symboltabelle in pi eingetragen, für die
		 * das nicht automatisch, während des parsens passiert ist.
		 */
		pi.setSaveProtocol(symbolTable._saveProtocol);

		if(makeProcessScript){
			if(symbolTable._pp.getProtocolType() == ProtocolType.Undefined) {
				// Der Benutzer gibt keinen expliziten Wert für die Protokollart an

				if(pi.getProtocolType() == ProtocolType.Undefined){
					// Falls auch im Skript nichts definiert wurde auf "Zustandsprotokoll" setzen
					pi.setProtocolType(ProtocolType.StatusProtocol);
				}

				// Ansonsten wurde der Wert in "pi" bereits beim Parsen gesetzt.
			}
			else {
				// Falls der Benutzer einen Protokolltyp angibt hat dieser immer Vorrang
				pi.setProtocolType(symbolTable._pp.getProtocolType());
			}

			// dasselbe mit der NoChangeMarker-information
			if(symbolTable._pp.getNoChangeMarker() == NoChangeMarker.Undefined) {
				// Der Benutzer gibt keinen expliziten Wert für die Protokollart an

				if(pi.getNoChangeMarker() == NoChangeMarker.Undefined){
					// Falls auch im Skript nichts definiert wurde auf "Zeile" setzen
					pi.setNoChangeMarker(NoChangeMarker.Row);
				}

				// Ansonsten wurde der Wert in "pi" bereits beim Parsen gesetzt.
			}
			else {
				// Falls der Benutzer einen Protokolltyp angibt hat dieser immer Vorrang
				pi.setNoChangeMarker(symbolTable._pp.getNoChangeMarker());
			}
		}

		if(symbolTable._periodList == null) {
			if(makeProcessScript) throw new SemanticErrorException(InterpreterErrorMessage.NO_PERIOD);
		}
		else {
			pi.setPeriods(symbolTable._periodList);
		}

		if(symbolTable._simVar == Qualifier.INVALID_SIMVAR) {
			if(makeProcessScript) throw new SemanticErrorException(InterpreterErrorMessage.NO_SIMVAR);
		}
		else {
			pi.setSimVar(symbolTable._simVar);
		}

		try {
			if(makeProcessScript) pi.setArchiveDataKinds(pp.getArchiveDataKindCombination());
		}
		catch(FailureException e) {
			throw new SemanticErrorException(e.getMessage());
		}

		pi.setMainObject(symbolTable.getMainObject());

		pi.setPseudoObjects(symbolTable._pseudoObjects);

		pi.setAspectBindings(symbolTable._aspectBindings);

		pi.setPossibleAspectBindings(symbolTable.getPossibleAspectBindings());

		pi.setAliases(symbolTable._aliases);

		for(final SemanticChecks sc : _semanticChecks) {
			sc.cleanUp(pp, makeProcessScript);
		}
		/*
		 * PI muss eigene Berechnungen durchführen, wie Liste der DIs erstellen!
		 * Wichtig: Vorher müssen die anderen Semantiküberprüfungen, insbesondere
		 * TempAttributeSemantics.cleanUp() (wg. Rekursionserkennung) aufgerufen werden!
		 */
		pi.update();

		//Überprüfen, ob zumindest EIN Zeitbereich existiert, falls das Protokoll erstellt werden soll
		if(makeProcessScript && (pi.getPeriods() == null || pi.getPeriods().size() == 0)) throw new SemanticErrorException(ErrorMessage.NO_PERIODS);
	}


	/**
	 * Gibt die zur Datenaufbereitung notwendigen Informationen zurück.
	 *
	 * @return ProcessingInformation Gibt die zur Datenaufbereitung notwendigen Informationen zurück.
	 */
	public ProcessingInformation getProcessingInformation() {
		return pi;
	}

	/**
	 * Liefert das Hilfsobjekt für Typüberprüfungen zurück.
	 *
	 * @return Hilfsobjekt für Typüberprüfungen.
	 */
	public MemberCheck getCheck() {
		return check;
	}

	/** @return Objekt das die Semantik von Aggregationen überprüft */
	public AggregationSemantics getAggregationSemantics() {
		return _aggregationSemantics;
	}

	/** @return Objekt das die Semantik von Attributgruppen überprüft */
	public AttributeGroupSemantics getAttributeGroupSemantics() {
		return _attributeGroupSemantics;
	}

	/** @return Objekt das die Semantik von Attributen aus dem Spaltenbereich überprüft */
	public AttributeSemantics getAttributeSemantics() {
		return _attributeSemantics;
	}

	/** @return Objekt das die Semantik von Ausdrücken überprüft */
	public ExpressionSemantics getExpressionSemantics() {
		return _expressionSemantics;
	}

	/** @return Objekt das die Semantik von diversen Konstrukten überprüft */
	public MiscellaneousSemantics getMiscSemantics() {
		return _miscSemantics;
	}

	/** @return Objekt das die Semantik von Realen Attributen überprüft */
	public RealAttributeSemantics getRealAttributeSemantics() {
		return _realAttributeSemantics;
	}

	/** @return Objekt das die Semantik des Mit-Konstrukts überprüft */
	public ScopeSemantics getScopeSemantics() {
		return _scopeSemantics;
	}

	/** @return Objekt das die Semantik von temporären Attributen überprüft */
	public TempAttributeSemantics getTempAttributeSemantics() {
		return _tempAttributeSemantics;
	}

	public SymbolTable getSymbolTable() {
		return symbolTable;
	}
}
