/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing.archivebuffer;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataQueryResult;
import de.bsvrz.pua.prot.processing.onlinebuffer.ExtResultData;
import de.bsvrz.sys.funclib.losb.datk.ProtocolType;
import de.bsvrz.pua.prot.util.Qualifier;
import de.bsvrz.pua.prot.util.attributes.RealAttributeDescription;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;

import java.util.Objects;

/**
 * Klasse die die Bestimmung des Archiv-Wertes eines Realen Attributes durchführt. Führt folgende Aufgaben durch:<br> - Archivanfragen stellen<br> -
 * Markierungsinformation erstellen<br> - Verschränken von Objekten<br> Wichtig: Objekte vom Typ {@link de.bsvrz.pua.prot.util.attributes.RealAttributeDescription}
 * und Objekte dieser Klasse liefern die gleichen Hashcodes und können auf Gleichheit geprüft werden. Siehe {@link #hashCode()} und {@link #equals(Object)}.
 * D.h. Objekte dieser beiden Klasse können als austauschbare Schlüssel für Hashtables / HashSets des Typs {@link de.bsvrz.pua.prot.util.attributes.RealElement}
 * verwendet werden.<br>
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class RealArchiveAttribute extends RealArchiveElement {

	/** Debug Ausgabe */
	private static final Debug _debug = Debug.getLogger();

	/** Beschreibung des Realen Attributs */
	private RealAttributeDescription _rad;

	/**
	 * @param archiveUser   Verbindung zum Archivsystem. Legt die Archivanfrageparameter fest.
	 * @param rad           Beschreibung des realen Attributs. Darf nicht null sein.
	 * @param protocolType  Protokollart
	 * @param isShown       Wird das Element im Spalten-Bereich verwendet? <code>true</code>Element wird im Spalten Bereich verwendet.
	 */
	public RealArchiveAttribute(ArchiveUser archiveUser, RealAttributeDescription rad, ProtocolType protocolType, boolean isShown) {
		super(archiveUser, rad.qualifier.getDataDescription(), rad.qualifier.object, protocolType, isShown, rad.padData());
		this._rad = rad;
	}

	/** @see RealArchiveElement#executeTake(long) */
	@Override
	protected boolean executeTake(long timeout) throws FailureException, InterruptedException {
		boolean result;

		/*
		 * Muss im synchronized block stehen, da this.archiveData nach außen NIE den vollständigen Datensatz speichern
		 * soll, den super.take() in resultData schreibt. sondern wirklich nur den Wert des Attributs Eigentlich wäre
		 * ein synchronized this.value ausreichend, aber das kann ja null sein
		 */
		synchronized(this) {
			// Daten holen
			result = super.executeTake(timeout);
			if(result && value != null) {
				if(_rad.getReference() != null) {
					getReferenceAndSelectData(_rad, value);
				}
				else {
					value.selectData(_rad);
				}
			}
		}
		return result;
	}

	/**
	 * Stellt eine Archivanfrage, um den Wert der Referenz herauszufinden.
	 *
	 * @param rad           Attributbeschreibung
	 * @param referenceData Datensatz, der das Objekt der Referenz enthält
	 */
	private void getReferenceAndSelectData(RealAttributeDescription rad, ExtResultData referenceData) throws InterruptedException {
		Data data = referenceData.getData();
		if(data != null) {
			long timeStamp = referenceData.getDataTime();
			if(timeStamp == 0) // Relativanfragen mit dem Zeitstempel 0 kann das Archivsystem nicht beantworten ->
			// Anfrage nach 1 stellen
			{
				timeStamp = 1;
			}
			ArchiveData ad = null;

			while(rad.getReference() != null && data != null && data.getReferenceValue(rad.getName()) != null) {
				Qualifier q = rad.getReference().qualifier.clone();
				q.object = data.getReferenceValue(rad.getName()).getSystemObject();
				ad = requestYounger(q, timeStamp);

				rad = rad.getReference();
				data = ad != null ? ad.getData() : null;
			}

			// Am Ende der Referenzkette angelangt
			if(ad != null) {
				this.value = new ExtResultData(ad, ad.getData(), ad.getDataIndex(), ad.getArchiveTime());
				// jetzt ggf noch die unwichtigen Daten verwerfen:
				value.selectData(rad);
			}
		}
	}

	/**
	 * Holt Archivdaten für die übergebene Qualifizierung. Dabei wird genau ein Wert aus dem Archivsystem geholt, und zwar derjenige, dessen Datenzeitstempel
	 * kleiner ist, als der übergebene Datenzeitstempel
	 *
	 * @param qualifier Qualifizierung
	 * @param timeStamp Datenzeitstempel
	 *
	 * @return Archivdatensatz oder null, falls es zu einem Fehler gekommen ist
	 */
	private ArchiveData requestYounger(Qualifier qualifier, long timeStamp) throws InterruptedException {
		ArchiveData result = null;
		try {
			ArchiveDataQueryResult qr = getArchiveUser().requestRelative(
					new DataDescription(qualifier.attributeGroup, qualifier.aspect), qualifier.object, timeStamp
			);
			if(qr.isRequestSuccessful()) {
				result = qr.getStreams()[0].take();
			}
			else {
				_debug.finest("Relative Archivanfrage wurde nicht durchgeführt: " + qr.getErrorMessage());
			}
		}
		catch(InterruptedException e) {
			// InterruptedException an aufrufende Methode weitergeben
			throw e;
		}
		catch(Exception e) {
			_debug.warning("Fehler bei der Archivanfrage: ", e);
		}
		return result;
	}

	/**
	 * Berechnet den Hash-Code. Der Hash-Code entspricht dem Hashcode von {@link #_rad}.
	 *
	 * @see java.lang.Object#hashCode()
	 */
	@Override
    public int hashCode() {
		return _rad.hashCode();
	}

	/**
	 * Vergleicht {@link #_rad}. Ist <code>other</code> vom Typ {@link RealArchiveAttribute} so werden die Variablen {@link #_rad} auf Gleichheit geprüft. Ist
	 * <code>other</code> vom Typ {@link de.bsvrz.pua.prot.util.attributes.RealAttributeDescription}, so wird überprüft, ob diese mit {@link #_rad} übereinstimmt.
	 *
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
    public boolean equals(Object other) {
		return Objects.equals(_rad, other);
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		return _rad + ": " + value;
	}
}
