/*
 * Copyright 2019 by Kappich Systemberatung Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.processing.onlinebuffer;

import de.bsvrz.dav.daf.accessControl.UserInfo;
import de.bsvrz.dav.daf.main.*;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.pua.prot.Protocols;
import de.bsvrz.sys.funclib.losb.exceptions.FailureException;

/**
 * Klasse dient zum Abfragen von aktuellen Datenwerten auf dem Dav. Wird für Referenzen verwendet. {@link #unsubscribe()} muss aufgerufen werden, nachdem die
 * Daten mit {@link #take(long)} abgerufen wurden.
 *
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
class ReferenceReceiver implements ClientReceiverInterface {

	/** Zeigt an ob Daten vorliegen. */
	private boolean receivedData = false;

	/** Daten. */
	private ResultData resultData = null;

	private final SystemObject object;

	private final DataDescription dataDescription;

	private final ClientDavInterface _clientDavInterface;

	/**
	 * Meldet den Empfang von Daten an.
	 *  @param systemObject   Objekt.
	 * @param attributeGroup Attributgruppe.
	 * @param aspect         Aspekt.
	 * @param options        Datenart.
*@param clientDavInterface
	 * @param userInfo
	 */
    public ReferenceReceiver(SystemObject systemObject, AttributeGroup attributeGroup, Aspect aspect, ReceiveOptions options, final ClientDavInterface clientDavInterface, final UserInfo userInfo) throws FailureException {
		this.object = systemObject;
	    this.dataDescription = new DataDescription(attributeGroup, aspect);

	    _clientDavInterface = clientDavInterface;

	    Protocols.checkPermissions(userInfo, object, dataDescription);
	    
	    _clientDavInterface.subscribeReceiver(this, systemObject, dataDescription, options, ReceiverRole.receiver());
	}

	/**
	 * Wird verwendet um Daten von Referenzen vom Dav abzuholen.
	 */
	public void update(ResultData[] results) {
		if(!receivedData) {
			resultData = results[results.length - 1];
			receivedData = true;
			synchronized(this) {
				notifyAll();
			}
		}
	}

	/**
	 * Meldet den Empfang von Daten ab.
	 *
	 */
    public void unsubscribe() {
	    _clientDavInterface.unsubscribeReceiver(this, object, dataDescription);
	}

	/**
	 * Gibt den aktuellen Wert der Datenidentifikation zurück. Meldet sich anschließend vom Datenverteiler ab.
	 *
	 * @param timeout Maximale Wartezeit.
	 *
	 * @return Aktueller Wert oder <code>null</code> falls keine Daten innerhalb der Wartezeit vorliegen.
	 *
	 * @throws InterruptedException Warten auf Antwortdatensatz wurde unterbrochen.
	 */
	public ResultData take(long timeout) throws InterruptedException {
		if(!receivedData) // keine while-Schleife, sonst wird der Timeout nicht berücksichtigt
		{
			synchronized(this) {
				wait(timeout);
			}
		}
		// TODO hier Zeitstempel der Referenz überprüfen!
		if(receivedData) {
			return resultData;
		}
		else {
			return null;
		}
	}

	/** @return Zeigt an ob Daten vorliegen. */
	public boolean hasData() {
		return receivedData;
	}
}
