/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 *
 * This file is part of de.bsvrz.pua.prot.
 *
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Contact Information:
 * Kappich Systemberatung
 * Martin-Luther-Straße 14
 * 52062 Aachen, Germany
 * phone: +49 241 4090 436
 * mail: <info@kappich.de>
 */

package de.bsvrz.pua.prot.util.attributes;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import de.bsvrz.pua.prot.util.Qualifier;

import java.util.*;

/**
 * Zeitdauer Attribut. Das Zeitdauer-Attribut wird wie ein reales Attribut verwendet. Der Wert des Attributs ist jedoch erst bekannt, nachdem alle realen
 * Attribute zweier aufeinanderfolgender Datensätze ausgewertet wurden. <br> Bitte beim Einsatz in Hashtables oder bei Vergleichen {@link #hashCode()} und
 * {@link #equals(Object)} beachten.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class DurationAttributeDescription extends AttributeDescription implements RealElement {

	/** Schlüsselwort des Zeitdauer-Attributs in der Gramatikdatei. */
	public static final String DURATION = "zeitdauer";

	/**
	 * Objekt das verwendet werden kann, um schnell auf Hashtables/Sets zuzugreifen.
	 *
	 * @see #equals(Object)
	 * @see #hashCode()
	 */
	private static final DurationAttributeDescription hashDad = new DurationAttributeDescription("Dummy");

	/** Name der Spalte. Wird durch das Schlüsselwort 'als' im Skript bestimmt. */
	public final String columnName;

	/** Zeitdauer-Attribut mir leerem Spaltennamen. */
	public DurationAttributeDescription() {
		this(null);
	}

	/**
	 * Zeitdauer-Attribut mit Spaltennamen.
	 *
	 * @param columnName Spaltenname des Zeitdauer-Attributs.
	 */
	public DurationAttributeDescription(String columnName) {
		this.columnName = columnName;
		this.setName(DURATION);
	}

	/**
	 * Liefert eine leere Collection zurück.
	 *
	 * @see de.bsvrz.pua.prot.util.attributes.Elements#getQualifiers()
	 */
	@Override
	public Collection<Qualifier> getQualifiers() {
		return ImmutableList.of();
	}

	/**
	 * Liefert ein leeres HashSet zurück.
	 *
	 * @see de.bsvrz.pua.prot.util.attributes.Elements#getRealElements()
	 */
	@Override
	public Set<? extends RealElement> getRealElements() {
		return ImmutableSet.of();
	}

	/** @see java.lang.Object#toString() */
	@Override
    public String toString() {
		if(columnName != null) {
			return "Zeitdauer als \"" + columnName + "\"";
		}
		else {
			return "Zeitdauer";
		}
	}

	/**
	 * Liefert immer 0 zurück.
	 *
	 * @see java.lang.Object#hashCode()
	 */
	@Override
    public int hashCode() {
		return 0;
	}

	/**
	 * Vergleicht zwei Objekte. Liefert <code>true</code> zurück, falls <code>obj</code> ein Objekt vom Typ {@link DurationAttributeDescription} ist.
	 *
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
    public boolean equals(Object obj) {
		if(obj instanceof DurationAttributeDescription) {
			return true;
		}
		else {
			return false;
		}
	}

	/**
	 * Liefert ein Objekt zurück, dass als Kex in Sets oder Hashtables verwendet werden kann.
	 *
	 * @return Zeitdauer-Attribut.
	 */
	public static DurationAttributeDescription forHash() {
		return hashDad;
	}

	/**
	 * Liefert immer <code>false</code> zurück.
	 *
	 * @see de.bsvrz.pua.prot.util.attributes.RealElement#padData()
	 */
	public boolean padData() {
		return false;
	}

	public int getIndex() {
		return -1;
	}
}
